# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

import sys, os, time, subprocess, platform
import signal
import atexit

pythonexe = "python3"
if platform.system() == "Windows":
    pythonexe = "python3w"

def kill_child(child_pid):
    if child_pid is None:
        pass
    else:
        os.kill(child_pid, signal.SIGTERM)

def start_script(local_script_name, auto_kill=True, log_file=None):
    global pythonexe
    
    if hasattr(sys, "frozen"):
        module_path = os.path.dirname(sys.executable)
    else:
        module_path = os.path.dirname(__file__)
    path = os.path.join(module_path, local_script_name)

    logging.info("Starting server using script: '%s', logging to '%s'" % (path, log_file))
    log_file = log_file if log_file is not None else os.devnull
    with open(log_file, 'w') as output:
        proc = subprocess.Popen([pythonexe, path], stdout=output)

    # Register this child pid to be killed when the parent dies:
    if auto_kill:
        atexit.register(kill_child, proc.pid)

    # Give it a sec
    time.sleep(1)



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .edit_mixture_controller import EditMixtureController
from .mixtures_controller import MixturesController

#from .edit_insitu_mixture_controller import EditInSituMixtureController
#from .insitu_behaviours_controller import InSituBehavioursController

#from .edit_insitu_behaviour_controller import EditInSituBehaviourController


__all__ = [
    #"EditInSituMixtureController",
    "EditMixtureController",
    "MixturesController",
    
    #"EditInSituBehaviourController"
    #"InSituBehavioursController",
]


#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon

import logging
logger = logging.getLogger(__name__)

from pyxrd.data.settings import ASYNC_SERER_PROVIDERS, ASYNC_SERVER_PRELOAD

_async_provider = None

def load_provider():
    global _async_provider
    logger.info("Loading async server provider")
    for class_path in ASYNC_SERER_PROVIDERS:
        logger.info("Trying to load async server provider at %s" % class_path)
        try:
            components = class_path.split('.')
            class_name = components[-1]
            mod = __import__('.'.join(components[:-1]), fromlist=[class_name])
            klass = getattr(mod, class_name)
        except (ImportError, AttributeError):
            logger.warning("Failed to import async provider %s!" % class_path)
            continue
        _async_provider = klass
        break
    logger.info("Loaded async server provider '%s'" % _async_provider)
    
# Pre-load async server if needed:
if ASYNC_SERVER_PRELOAD:
    load_provider()
    if _async_provider is not None:
        _async_provider.launch_server()
    
def get_provider():
    global _async_provider
    if _async_provider is None:
        load_provider()
    return _async_provider

def get_status():
    return get_provider().get_status()


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import numpy as np
from math import radians, tan
from scipy.interpolate.interpolate import interp1d

from .goniometer import get_fixed_to_ads_correction_range
from .phases import get_intensity


def get_clipped_intensities(specimen):
    exp = specimen.observed_intensity.transpose()[...,specimen.selected_range]
    cal = specimen.total_intensity[...,specimen.selected_range]
    return exp, cal

def get_machine_correction_range(specimen):
    """
        Calculate a correction factor for a certain sample length,
        sample absorption and machine setup.
    """
    goniometer = specimen.goniometer
    range_st = np.sin(specimen.range_theta)
    correction_range = np.ones_like(specimen.range_theta)
    
    # Correct for automatic divergence slits first:
    if goniometer.divergence_mode == "AUTOMATIC":
        correction_range *= get_fixed_to_ads_correction_range(
            specimen.range_theta, goniometer)
    # Then correct for sample absorption:
    if goniometer.has_absorption_correction > 0.0:
        absorption = goniometer.absorption * goniometer.sample_surf_density * 1e-3
        if absorption > 0.0:
            correction_range *= np.minimum(1.0 - np.exp(-2.0 * absorption / range_st), 1.0)
    # And finally correct for sample length, if fixed slits were used:
    if goniometer.divergence_mode == "FIXED" and goniometer.divergence > 0:
        L_Rta = goniometer.sample_length / (goniometer.radius * tan(radians(goniometer.divergence)))
        correction_range *= np.minimum(range_st * L_Rta, 1)
    return correction_range

def calculate_phase_intensities(specimen):
    """
        Gets phase intensities for the provided phases
        Returns a 2-tuple containing 2-theta values and phase intensities.
    """
    range_stl = 2 * np.sin(specimen.range_theta) / specimen.goniometer.wavelength

    correction_range = get_machine_correction_range(specimen)

    def get_phase_intensities(phases):
        for phase in phases:
            if phase != None:
                correction = correction_range if phase.apply_correction else 1.0
                yield get_intensity(
                    specimen.range_theta, range_stl,
                    specimen.goniometer.soller1, specimen.goniometer.soller2, specimen.goniometer.mcr_2theta,
                    phase
                ) * correction
            else:
                yield np.zeros_like(range_stl)

    def interpolate_wavelength(I, new_wavelength, fraction):
        new_theta = np.arcsin(range_stl * new_wavelength * 0.5)
        f = interp1d(new_theta, I * fraction, bounds_error=False, fill_value=0.0)
        return f(specimen.range_theta)

    def apply_wavelength_distribution(I):
        for new_wavelength, fraction in specimen.goniometer.wavelength_distribution:
            I += interpolate_wavelength(I, new_wavelength, fraction)
        return I

    return (
        correction_range,
        np.swapaxes(np.array([
            [ apply_wavelength_distribution(I) for I in get_phase_intensities(specimen.phases[z_index]) ]
              for z_index in range(len(specimen.z_list))
        ], dtype=np.float_), 0, 1) # is of shape num_phases, num_patterns, num_observations
     )

def calculate_scaled_intensities(specimen, scale, fractions, bgshift):
    specimen.background_intensity = bgshift * specimen.correction
    specimen.scaled_phase_intensities = (fractions * specimen.phase_intensities.transpose()).transpose() * scale
    specimen.total_intensity = np.sum(specimen.scaled_phase_intensities, axis=0) + specimen.background_intensity
    return specimen



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.


class Cancellable(object):
    """
        Object which has a (threaded) action that can be cancelled by the user.
    """
    _stop = None

    def _user_cancelled(self):
        return bool(self._stop is not None and self._stop.is_set())

    pass #end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#  Copyright (C) 2005 by Tobias Weber
#  Copyright (C) 2005 by Roberto Cavada <roboogle@gmail.com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

from ..decorators import good_classmethod_decorator
from .base import ObsWrapperBase

class Observable (ObsWrapperBase):

    @classmethod
    @good_classmethod_decorator
    def observed(cls, _func):
        """
        Decorate methods to be observable. If they are called on an instance
        stored in a property, the model will emit before and after
        notifications.
        """

        def wrapper(*args, **kwargs):
            self = args[0]
            assert(isinstance(self, Observable))

            self._notify_method_before(self, _func.__name__, args, kwargs)
            res = _func(*args, **kwargs)
            self._notify_method_after(self, _func.__name__, res, args, kwargs)
            return res

        return wrapper

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

from traceback import print_exc
import logging
logger = logging.getLogger(__name__)

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk, GObject  # @UnresolvedImport

from .base_models import BaseObjectListStore
from ....observers import TreeObserver

class ObjectTreeStore(BaseObjectListStore):
    """
        GenericTreeModel implementation that holds a tree with objects.
        It expects all objects to be of a certain type, which needs to be
        passed to the __init__ as the first argument. 
    """

    # PROPERTIES:
    _object_node_map = None

    @property
    def _root_node(self):
        return getattr(self._model, self._prop_name, None)

    def is_wrapping(self, model, prop_name):
        return self._model == model and self._prop_name == prop_name

    # ------------------------------------------------------------
    #      Initialization and other internals
    # ------------------------------------------------------------
    def __init__(self, model, prop):
        _root = getattr(model, prop.label, None)

        # Then continue:
        try:
            BaseObjectListStore.__init__(self, prop.data_type)
        except ValueError as err:
            msg = "ValueError (%r) was raised when initializing ObjectTreeStore for model '%s' and data type '%s'" % (err, model, prop.data_type)
            msg += "\n Did you forget to set the data_type on the list property '%s'?" % prop.label
            raise ValueError(msg)
        self._model = model
        self._prop_name = prop.label
        self._object_node_map = dict()

        self._observer = TreeObserver(
            self.on_item_inserted, self.on_item_deleted,
            on_deleted_before=self.on_item_deleted_before,
            prop_name=self._prop_name, model=self._model
        )

    def on_item_inserted(self, item):
        try:
            itr = self.create_tree_iter(item)
            path = self.get_path(itr)
            # self._observe_item(item)
            self.row_inserted(path, itr)
        except ValueError:
            logger.debug("Invalid rowref passed: %s", item)
            pass # invalid rowref

    _deleted_paths = []
    def on_item_deleted_before(self, item):
        # self._unobserve_item(item)
        self._deleted_paths.append(self.on_get_path(item))

    def on_item_deleted(self, item):
        for path in self._deleted_paths:
            self.row_deleted(path)
        self._deleted_paths = []

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def on_get_flags(self):
        return Gtk.TreeModelFlags.ITERS_PERSIST

    def on_get_iter(self, path):
        try:
            if hasattr(path, 'split'): path = list(map(int, path.split(":")))
            return self._root_node.get_child_node(*path)
        except IndexError as err:
            err.args = "IndexError in on_get_iter of %s caused by %s" % (self, path)
            print_exc()
            return None

    def on_get_path(self, node):
        try:
            return ":".join(map(str, node.get_indices()))
        except ValueError as err:
            err.args = "ValueError in on_get_path of %s caused by %s" % (self, node)
            print_exc()
            return None

    def set_value(self, itr, column, value):
        user_data = self.get_tree_node_object(itr)
        setattr(user_data, self._columns[column][0], value)
        self.row_changed(self.get_path(itr), itr)

    def on_get_value(self, node, column):
        try:
            return getattr(node.object, self._columns[column][0])
        except:
            return ""

    def on_iter_next(self, node):
        return node.get_next_node()

    def on_iter_children(self, node):
        node = node or self._root_node
        return node.get_first_child_node()

    def on_iter_has_child(self, node):
        node = node or self._root_node
        return node.has_children

    def on_iter_n_children(self, node):
        node = node or self._root_node
        return node.child_count

    def on_iter_nth_child(self, parent, n):
        node = parent or self._root_node
        try:
            return node.get_child_node(n)
        except:
            return None

    def on_iter_parent(self, node):
        return node.parent

    def iter_objects(self):
        for node in self._root_node.iter_children():
            yield node.object

    def get_tree_node(self, itr):
        return BaseObjectListStore.get_user_data(self, itr)

    def get_tree_node_from_path(self, path):
        return BaseObjectListStore.get_user_data_from_path(self, path)

    def get_tree_node_object(self, itr):
        return self.get_tree_node(itr).object

    def get_tree_node_object_from_path(self, path):
        return self.get_tree_node_from_path(path).object

    pass # end of class

GObject.type_register(ObjectTreeStore) # @UndefinedVariable



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

import random, copy

import numpy as np

from deap.tools import HallOfFame
from deap import creator, base, tools #@UnresolvedImport

from pyxrd.generic.asynchronous.cancellable import Cancellable
from pyxrd.refinement.refine_async_helper import RefineAsyncHelper

from ..refine_method import RefineMethod
from ..refine_method_option import RefineMethodOption

from .deap_utils import pyxrd_array, FitnessMin, result_func
from .deap_cma import Strategy

# Default settings:
NGEN = 100
NGEN_COMM = 5
NSWARMS = 4

class SwarmStrategy(Cancellable):

    def __create_strategy(self, parent, sigma, ranges, **kwargs):
        return Strategy(
            centroid=parent, sigma=sigma, ranges=ranges,
            stop=self._stop, **kwargs
        )

    def __init__(self, parents, sigma, ranges, stop, ** kwargs):
        self.nswarms = len(parents)
        self._stop = stop
        self.strategies = [self.__create_strategy(parents[i], sigma, ranges, **kwargs) for i in range(self.nswarms)]
        self.global_best = None

    def update(self, swarms, communicate=False):
        if self._user_cancelled():
            return

        for i, population in enumerate(swarms):
            self.strategies[i].update(population)
            # Keep track of the global best:
            best = population[0]
            try:
                if self.global_best is None or np.all(self.global_best.fitness < best.fitness):
                    self.global_best = copy.deepcopy(best)
            except ValueError:
                logger.warn("Got a value error comparing '%s' and '%s'" % (self.global_best.fitness, best.fitness))

        if communicate:
            for i, population in enumerate(swarms):
                self.strategies[i].rotate_and_bias(self.global_best)

    def generate(self, ind_init):
        for strategy in self.strategies:
            yield strategy.generate(ind_init)

    pass #end of class


class SwarmAlgorithm(RefineAsyncHelper):

    @property
    def ngen(self):
        return self._ngen
    @ngen.setter
    def ngen(self, value):
        self._ngen = value
        logger.info("Setting ngen to %d" % value)
    _ngen = 100

    gen = -1
    halloffame = None

    refiner = None

    toolbox = None
    stats = None
    verbose = False

    #--------------------------------------------------------------------------
    #    Initialization
    #--------------------------------------------------------------------------
    def __init__(self, toolbox, halloffame, stats, ngen=NGEN, ngen_comm=NGEN_COMM,
                 verbose=__debug__, refiner=None, stop=None, eval_func=None, result_func=None):
        """
        :param toolbox: A :class:`~deap.base.Toolbox` that contains the evolution
                        operators.
        :param ngen: The number of generations.
        :param ngen_comm: At each multiple generation of this number swarms will
                         communicate
        :param halloffame: A :class:`~deap.tools.ParetoFront` object that will
                           contain the best individuals.
        :param stats: A :class:`~deap.tools.Statistics` object that is updated
                      inplace.
        :param verbose: Whether or not to log the statistics.
                    
        :param refiner: PyXRD refiner object
    
        :returns: The best individual and the final population.
        
        The toolbox should contain a reference to the generate and the update method 
        of the chosen strategy.
        
        Call the run() method when the algorithm should be run.
    
        .. [Colette2010] Collette, Y., N. Hansen, G. Pujol, D. Salazar Aponte and
           R. Le Riche (2010). On Object-Oriented Programming of Optimizers -
           Examples in Scilab. In P. Breitkopf and R. F. Coelho, eds.:
           Multidisciplinary Design Optimization in Computational Mechanics,
           Wiley, pp. 527-565;
    
        """
        self.stats = stats
        self.toolbox = toolbox
        self.ngen = ngen
        self.ngen_comm = ngen_comm
        self.halloffame = halloffame
        self.verbose = verbose
        self.refiner = refiner

        self.eval_func = eval_func
        self.result_func = result_func

        self.gen = 0

        self._stop = stop

    #--------------------------------------------------------------------------
    #    Run method:
    #--------------------------------------------------------------------------
    def run(self):
        """Will run this algorithm"""
        if self.verbose:
            column_names = ["gen", "evals", "best"]
            if self.stats is not None:
                column_names += list(self.stats.functions.keys())
            self.logbook = tools.Logbook()
            self.logbook.header = column_names

        for _ in range(self.ngen):
            # Check if the user has cancelled:
            if self._user_cancelled():
                self.refiner.status.message = "Stopping..."
                logger.info("User cancelled execution, stopping ...")
                break

            #ASK: Generate a new population:
            swarms = self._ask()
            #TELL: Update the strategy with the evaluated individuals
            self._tell(swarms)
            #RECORD: For logging:
            self._record(swarms)

        return self.refiner.history.best_solution, [ind for population in swarms for ind in population]

    #--------------------------------------------------------------------------
    #    Ask, tell & record:
    #--------------------------------------------------------------------------
    def _ask(self):
        self.gen += 1
        self.refiner.status.message = "Creating generation #%d" % self.gen

        swarms = []
        def iter_func():
            for generator in self.toolbox.generate():
                swarm = []
                for solution in generator:
                    swarm.append(solution)
                    yield solution
                swarms.append(swarm)

        if self.result_func is None:
            self.result_func = result_func
            
        def result_f(*args):
            self.refiner.update(*args)
            self.result_func(*args)

        population = self.do_async_evaluation(
            iter_func=iter_func,
            eval_func=self.eval_func,
            result_func=result_f
        )

        if self.halloffame is not None:
            self.halloffame.update(population)

        del population

        return swarms

    def _tell(self, swarms):
        self.refiner.status.message = "Updating strategy"
        communicate = bool(self.gen > 0 and self.gen % self.ngen_comm == 0)
        self.toolbox.update(swarms, communicate=communicate)

    def _record(self, swarms):
        self.refiner.status.message = "Processing ..."

        # Get the best solution so far:
        if hasattr(self.halloffame, "get_best"):
            best = self.halloffame.get_best()
        else:
            best = self.halloffame[0]
        best_f = best.fitness.values[0]

        flat_pop = [ind for population in swarms for ind in population]
        pop_size = len(flat_pop)

        # Calculate stats & print something if needed:
        record = self.stats.compile(flat_pop)
        if self.verbose:
            self.logbook.record(gen=self.gen, evals=pop_size, best=best_f, **record)
            print(self.logbook.stream)


        self.refiner.status.message = "Refiner update ..."
        # Update the refiner history:
        self.refiner.update(best, iteration=self.gen, residual=best_f)

    pass #end of class

class RefinePSOCMAESRun(RefineMethod):
    """
        The PS-CMA-ES hybrid algorithm implementation
    """
    name = "PS-CMA-ES refinement"
    description = "This algorithm uses the PS-CMA-ES hybrid refinement strategy"
    index = 6
    disabled = False

    ngen = RefineMethodOption('Maximum # of generations', NGEN, [1, 10000], int)
    nswarms = RefineMethodOption('# of CMA swarms', NSWARMS, [1, 100], int)
    ngen_comm = RefineMethodOption('Communicate each x gens', NGEN_COMM, [1, 10000], int)

    def _individual_creator(self, refiner, bounds):
        creator.create(
            "Individual", pyxrd_array,
            fitness=FitnessMin, # @UndefinedVariable
            refiner=refiner,
            min_bounds=bounds[:, 0].copy(),
            max_bounds=bounds[:, 1].copy(),
        )

        def create_individual(lst):
            arr = np.array(lst).clip(bounds[:, 0], bounds[:, 1]) #@UndefinedVariable
            return creator.Individual(arr) # @UndefinedVariable

        return create_individual

    def _create_stats(self):
        stats = tools.Statistics(lambda ind: ind.fitness.values)
        stats.register("avg", np.mean, axis=0) #@UndefinedVariable
        stats.register("std", np.std, axis=0) #@UndefinedVariable
        stats.register("min", np.min, axis=0) #@UndefinedVariable
        stats.register("max", np.max, axis=0) #@UndefinedVariable
        return stats

    _has_been_setup = False
    def _setup(self, refiner, ngen=NGEN, ngen_comm=NGEN_COMM, nswarms=NSWARMS, **kwargs):
        if not self._has_been_setup:
            logger.info("Setting up the DEAP CMA-ES refinement algorithm (ngen=%d)" % ngen)
            refiner.status.message = "Setting up algorithm..."

            # Process some general stuff:
            bounds = np.array(refiner.ranges) #@UndefinedVariable
            create_individual = self._individual_creator(refiner, bounds)

            # Setup strategy:

            #TODO make the others random
            parents = [None] * nswarms
            parents[0] = create_individual(refiner.history.initial_solution)

            for i in range(1, nswarms):
                parents[i] = create_individual([
                   random.uniform(bounds[j, 0], bounds[j, 1]) for j in range(len(refiner.history.initial_solution))
                ])

            strategy = SwarmStrategy(
                parents=parents, sigma=1.0 / 10.0, ranges=bounds,
                stop=self._stop,
            )

            # Toolbox setup:
            toolbox = base.Toolbox()
            toolbox.register("generate", strategy.generate, create_individual)
            toolbox.register("update", strategy.update)

            # Hall of fame & stats:
            logger.info("Creating hall-off-fame and statistics")
            halloffame = HallOfFame(1, similar=lambda a1, a2: np.all(a1 == a2)) #PyXRDParetoFront(similar=lambda a1, a2: np.all(a1 == a2))
            stats = self._create_stats()

            # Create algorithm
            self.algorithm = SwarmAlgorithm(
                toolbox, halloffame, stats, ngen=ngen, ngen_comm=ngen_comm,
                refiner=refiner, stop=self._stop)

            self._has_been_setup = True
        return self.algorithm

    def run(self, refiner, **kwargs):
        logger.info("CMA-ES run invoked with %s" % kwargs)
        self._has_been_setup = False #clear this for a new refinement
        algorithm = self._setup(refiner, **kwargs)
        # Get this show on the road:
        logger.info("Running the CMA-ES algorithm...")
        algorithm.run()

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from mvc.support.gui_loop import add_idle_call

class FloatEntryValidator:
    def __init__(self, entry):
        self.last_valid_val = 0
        self.has_valid_val = True
    
        self.entry = entry
        self.entry.connect("activate", self.entry_activate)
        self.entry.connect("focus_out_event", self.entry_focus_out)
        self.insert_handlerid = self.entry.connect("insert-text", self.entry_insert_text)
        self.delete_handlerid = self.entry.connect("delete-text", self.entry_delete_text)

    def validate(self, text=None, reset_if_invalid=False):
        text = text or self.entry.get_chars(0, -1)
        try:
            self.last_valid_val = float(text)
            self.has_valid_val = True
        except Exception as e:
            self.has_valid_val = False
        if reset_if_invalid and not self.has_valid_val:
            self.entry.handler_block(self.insert_handlerid)
            self.entry.set_text("%f" % self.last_valid_val)
            self.has_valid_val = True
            self.entry.handler_unblock(self.insert_handlerid)

    def entry_activate(self, entry):
        self.validate(reset_if_invalid=True)

    def entry_focus_out(self, entry, event):
        self.validate(reset_if_invalid=True)
        return False

    def entry_insert_text(self, entry, new_text, new_text_length, position):
        self.entry.stop_emission('insert-text')
        self.entry.handler_block(self.insert_handlerid)
        pos = self.entry.get_position()
        
        text = self.entry.get_chars(0, -1)
        old_text = text
        
        text = text[:pos] + new_text + text[pos:]
        
        self.validate(text)
        if self.has_valid_val:
            new_text = text
            self.entry.set_text(new_text)
            add_idle_call(lambda: self.entry.set_position(pos + (len(new_text) - len(old_text))))
        #while Gtk.events_pending():
        #    Gtk.main_iteration(False)
        self.entry.handler_unblock(self.insert_handlerid)        
        
    def entry_delete_text(self, entry, start, end):
        self.validate()



# coding=UTF-8
# ex:ts=4:sw=4:et:
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import Gtk, GObject, GLib  # @UnresolvedImport

from mvc.support.utils import round_sig

class ScaleEntry(Gtk.HBox):
    """
        The ScaleEntry combines the generic GtkEntry and GtkScale widgets in
        one widget, with synchronized values and one changed signal.
    """

    __gsignals__ = {
        'changed' : (GObject.SignalFlags.RUN_LAST, None, []), #@UndefinedVariable
    }

    @property
    def lower(self):
        return self.adjustment.get_lower()
    @lower.setter
    def lower(self, value):
        return self.adjustment.set_lower(value)

    @property
    def upper(self):
        return self.adjustment.get_upper()
    @upper.setter
    def upper(self, value):
        return self.adjustment.set_upper(value)

    def __init__(self, lower=0, upper=10, enforce_range=False):
        GObject.GObject.__init__(self, spacing=5)

        self.enforce_range = enforce_range

        if lower == None: lower = 0
        if upper == None: upper = 10
        lower = min(upper, lower)
        upper = max(upper, lower)

        step = max((upper - lower) / 200.0, 0.01)
        self.adjustment = Gtk.Adjustment(
            0.0, lower, upper, step, step, 0.0)

        self.adjustment.connect('value-changed', self.on_adj_value_changed)

        self.scale = Gtk.Scale.new(Gtk.Orientation.HORIZONTAL, self.adjustment)
        self.scale.set_draw_value(False)
        self.scale.set_size_request(50, -1)

        self.entry = Gtk.SpinButton()
        self.entry.set_adjustment(self.adjustment)
        self.entry.set_digits(5)
        self.entry.set_numeric(True)
        self.entry.set_size_request(150, -1)

        self.set_value(self.scale.get_value())

        Gtk.HBox.pack_start(self, self.scale, False, True, 0)
        Gtk.HBox.pack_start(self, self.entry, False, True, 0)
        self.set_focus_chain((self.entry,))


    _idle_changed_id = None
    def _idle_emit_changed(self):
        if self._idle_changed_id is not None:
            GLib.source_remove(self._idle_changed_id)
        self._idle_changed_id = GLib.idle_add(self._emit_changed)

    def _emit_changed(self):
        self.emit('changed')
        self._idle_changed_id = None;

    def on_adj_value_changed(self, adj, *args):
        self._idle_emit_changed()

    def _update_adjustment(self, lower, upper):
        step = round_sig(max((upper - lower) / 200.0, 0.0005))
        self.adjustment.configure(lower, upper,
            step, step, 0.0)

    def _update_range(self, value):
        lower, upper = self.lower, self.upper
        if not self.enforce_range:
            if value < (lower + abs(lower) * 0.05):
                lower = value - abs(value) * 0.2
            if value > (upper - abs(lower) * 0.05):
                upper = value + abs(value) * 0.2
            self._update_adjustment(lower, upper)

    def set_value(self, value):
        self._update_range(value)
        self.adjustment.set_value(value)

    def get_value(self):
        return self.adjustment.get_value()

    def get_children(self, *args, **kwargs):
        return []
    def add(self, *args, **kwargs):
        raise NotImplementedError
    def add_with_properties(self, *args, **kwargs):
        raise NotImplementedError
    def child_set(self, *args, **kwargs):
        raise NotImplementedError
    def child_get(self, *args, **kwargs):
        raise NotImplementedError
    def child_set_property(self, *args, **kwargs):
        raise NotImplementedError
    def child_get_property(self, *args, **kwargs):
        raise NotImplementedError
    def remove(self, *args, **kwargs):
        raise NotImplementedError
    def set_child_packing(self, *args, **kwargs):
        raise NotImplementedError
    def query_child_packing(self, *args, **kwargs):
        raise NotImplementedError
    def reorder_child(self, *args, **kwargs):
        raise NotImplementedError
    def pack_start(self, *args, **kwargs):
        raise NotImplementedError
    def pack_end(self, *args, **kwargs):
        raise NotImplementedError

    pass # end of class

GObject.type_register(ScaleEntry) #@UndefinedVariable



__add_idle_call = None
__remove_idle_call = None
__add_timeout_call = None
__remove_timeout_call = None
__setup_event_loop = None
__start_event_loop = None
__stop_event_loop = None

__idle_call_dict = {}
__timeout_call_dict = {}

def add_idle_call(func, *args, **kwargs):
    global __add_idle_call
    global __idle_call_dict
    if __add_idle_call is not None:
        __idle_call_dict[func] = __add_idle_call(func, *args, **kwargs)
    else: # toolkit does not support this or is not loaded:
        func(*args, **kwargs)
        
def remove_idle_call(func):
    global __remove_idle_call
    global __idle_call_dict
    if __remove_idle_call is not None:
        __remove_idle_call(__idle_call_dict[func])
        
def add_timeout_call(timeout, func, *args, **kwargs):
    global __add_timeout_call
    global __timeout_call_dict
    if __add_timeout_call is not None:
        __timeout_call_dict[func] = __add_timeout_call(timeout, func, *args, **kwargs)
    else: # toolkit does not support this or is not loaded:
        func(*args, **kwargs)

def remove_timeout_call(func):
    global __remove_timeout_call
    global __timeout_call_dict
    if __remove_timeout_call is not None:
        __remove_timeout_call(__timeout_call_dict[func])

def start_event_loop():
    global __start_event_loop
    if __start_event_loop is not None:
        return __start_event_loop()

def stop_event_loop():
    global __stop_event_loop
    if __stop_event_loop is not None:
        return __stop_event_loop()    


def load_toolkit_functions(
        add_idle_call, 
        remove_idle_call,
        add_timeout_call,
        remove_timeout_call,
        start_event_loop,
        stop_event_loop):
    """
        'add_idle_call' should take a function as 1st argument, the return 
        value is passed back to 'remove_idle_call'. Internally a cache is maintained
        in which keys are functions and values are return values.
        'add_timeout_call' and 'remove_timeout_call' work analogously
        start_event_loop and stop_event_loop don't take arguments and should be self-explanatory.
    """
    global __add_idle_call
    global __remove_idle_call
    global __add_timeout_call
    global __remove_timeout_call
    global __start_event_loop
    global __stop_event_loop
    assert callable(add_idle_call)
    assert callable(remove_idle_call)
    assert callable(add_timeout_call)
    assert callable(remove_timeout_call)
    assert callable(start_event_loop)
    assert callable(stop_event_loop)
    __add_idle_call = add_idle_call
    __remove_idle_call = remove_idle_call
    __add_timeout_call = add_timeout_call
    __remove_timeout_call = remove_timeout_call
    __start_event_loop = start_event_loop
    __stop_event_loop = stop_event_loop

"""
    Decorators:
"""

def run_when_idle(func):
    def callback(*args, **kwargs):
        return add_idle_call(func, *args, **kwargs)
    return callback
    
def run_every(timeout):
    def wrapper(func):
        def callback(*args, **kwargs):
            return add_timeout_call(func, timeout, *args, **kwargs)
        return callback
    return wrapper



import json
from datetime import datetime
from threading import Thread
from warnings import warn

from . import utils
from .mywarnings import *

__version__ = '0.1.2'


def check_outdated(package, version):
    """
    Given the name of a package on PyPI and a version (both strings), checks
    if the given version is the latest version of the package available.

    Returns a 2-tuple (is_outdated, latest_version) where
    is_outdated is a boolean which is True if the given version is earlier
    than the latest version, which is the string latest_version.

    Attempts to cache on disk the HTTP call it makes for 24 hours. If this
    somehow fails the exception is converted to a warning (OutdatedCacheFailedWarning)
    and the function continues normally.
    """

    from pkg_resources import parse_version

    parsed_version = parse_version(version)
    latest = None

    with utils.cache_file(package, 'r') as f:
        content = f.read()
        if content:  # in case cache_file fails and so f is a dummy file
            latest, cache_dt = json.loads(content)
            if not utils.cache_is_valid(cache_dt):
                latest = None

    def get_latest():
        url = 'https://pypi.python.org/pypi/%s/json' % package
        response = utils.get_url(url)
        return json.loads(response)['info']['version']

    if latest is None:
        latest = get_latest()

    parsed_latest = parse_version(latest)

    if parsed_version > parsed_latest:

        # Probably a stale cached value
        latest = get_latest()
        parsed_latest = parse_version(latest)

		# Don't be stupid - I'm building more recent version locally that have not been released yet
        #if parsed_version > parsed_latest:
        #    raise ValueError('Version %s is greater than the latest version on PyPI: %s' %
        #                     (version, latest))

    is_latest = parsed_version == parsed_latest
    #assert is_latest or parsed_version < parsed_latest again don't be stupid

    with utils.cache_file(package, 'w') as f:
        data = [latest, utils.format_date(datetime.now())]
        json.dump(data, f)

    return not is_latest, latest


def warn_if_outdated(package,
                     version,
                     raise_exceptions=False,
                     background=True,
                     ):
    """
    Higher level convenience function using check_outdated.

    The package and version arguments are the same.

    If the package is outdated, a warning (OutdatedPackageWarning) will
    be emitted.

    Any exception in check_outdated will be converted to a warning (OutdatedCheckFailedWarning)
    unless raise_exceptions if True.

    If background is True (the default), the check will run in
    a background thread so this function will return immediately.
    In this case if an exception is raised and raise_exceptions if True
    the traceback will be printed to stderr but the program will not be
    interrupted.

    This function doesn't return anything.
    """

    def check():
        # noinspection PyUnusedLocal
        is_outdated = False
        with utils.exception_to_warning('check for latest version of %s' % package,
                                        OutdatedCheckFailedWarning,
                                        always_raise=raise_exceptions):
            is_outdated, latest = check_outdated(package, version)

        if is_outdated:
            warn('The package %s is out of date. Your version is %s, the latest is %s.'
                 % (package, version, latest),
                 OutdatedPackageWarning)

    if background:
        thread = Thread(target=check)
        thread.start()
    else:
        check()


warn_if_outdated('outdated', __version__)



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pkg_resources import resource_filename # @UnresolvedImport

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

from mvc.adapters.gtk_support.widgets import ScaleEntry
from mvc.support.utils import rec_getattr

from pyxrd.generic.views import BaseView, HasChildView
from pyxrd.probabilities.models import RGbounds

def get_correct_probability_views(probability, parent_view):
    """
        Convenience function that creates both an `IndependentsView` and 
        `MatrixView` based on the probability model passed.
    """
    if probability is not None:
        G = probability.G
        R = probability.R
        rank = probability.rank
        if (RGbounds[R, G - 1] > 0):
            return IndependentsView(meta=probability.Meta, parent=parent_view), MatrixView(R=R, G=G, rank=rank, parent=parent_view)
        else:
            raise ValueError("Cannot (yet) handle R%d for %d layer structures!" % (R, G))

class EditProbabilitiesView(HasChildView, BaseView):
    """
        Container view containing one `MatrixView` and one `IndependentsView`
    """
    builder = resource_filename(__name__, "glade/probabilities.glade")
    top = "edit_probabilities"

    independents_container = "independents_box"
    independents_view = None
    dependents_container = "dependents_box"
    dependents_view = None

    widget_format = "prob_%s"

    def set_views(self, independents_view, dependents_view):
        self.independents_view = independents_view
        self._add_child_view(independents_view.get_top_widget(), self[self.independents_container])
        self.dependents_view = dependents_view
        self._add_child_view(dependents_view.get_top_widget(), self[self.dependents_container])
        self.show_all()
        return self.independents_view, self.dependents_view

class ProbabilityViewMixin():
    """
        Mixin class providing interface code for controllers
        of both `MatrixView` and `IndependentsView`
    """
    def update_matrices(self, W, P):
        raise NotImplementedError

class IndependentsView(HasChildView, ProbabilityViewMixin, BaseView):
    """
        Generic view that is able to generate an two-column list of inputs and 
        labels using the models Meta (passed to the constructor).
    """
    builder = resource_filename(__name__, "glade/R0_independents.glade")
    top = "R0independents_box"
    # generated table of weight fractions! (split in two columns)

    lbl_widget = "lbl_independents"
    sep_widget = "seperator_i"

    widget_format = "prob_%s"

    def __init__(self, meta, **kwargs):
        assert (meta is not None), "IndependentsView needs a model's Meta class!"
        BaseView.__init__(self, **kwargs)

        self.props = [ prop for prop in meta.all_properties if getattr(prop, "is_independent", False) ]
        all_props = { prop.label: prop for prop in meta.all_properties }

        N = len(self.props)

        def create_inputs(table):
            input_widgets = [None] * N
            check_widgets = [None] * N


            num_columns = 2
            column_width = 3

            for i, prop in enumerate(self.props):

                new_lbl = self.create_mathtext_widget(prop.math_title, prop.label)

                new_inp = ScaleEntry(lower=prop.minimum, upper=prop.maximum, enforce_range=True)
                new_inp.set_tooltip_text(prop.title)
                new_inp.set_name(self.widget_format % prop.label)
                self[self.widget_format % prop.label] = new_inp
                input_widgets[i] = new_inp

                j = (i % num_columns) * column_width
                table.attach(new_lbl, 0 + j, 1 + j, i / num_columns, (i / num_columns) + 1, xpadding=2, ypadding=2)
                table.attach(new_inp, 2 + j, 3 + j, i / num_columns, (i / num_columns) + 1, xpadding=2, ypadding=2)

                if prop.inheritable is not None:
                    inh_prop = all_props.get(prop.inherit_flag, None)
                    if inh_prop is None:
                        raise ValueError("The inherit flag property `%s` is missing for `%s` on meta model with store id `%s`" % (
                            prop.inherit_flag,
                            prop.label,
                            meta.store_id
                        ))

                    new_check = Gtk.CheckButton(label="")
                    new_check.set_tooltip_text(inh_prop.title)
                    new_check.set_name(self.widget_format % inh_prop.label)
                    new_check.set_sensitive(False)
                    self[self.widget_format % inh_prop.label] = new_check
                    check_widgets[i] = new_check
                    table.attach(new_check, 1 + j, 2 + j, i / num_columns, (i / num_columns) + 1, xpadding=2, ypadding=2, xoptions=Gtk.AttachOptions.FILL)


                del new_inp, new_lbl
            return input_widgets, check_widgets
        self.i_box = self['i_box']

        num_rows = int((N + 1) / 2)
        if not num_rows == 0:
            self.i_table = Gtk.Table(num_rows, 4, False)
            self.i_inputs, self.i_checks = create_inputs(self.i_table)
        else:
            self.i_inputs, self.i_checks = [], []
        if len(self.i_inputs) == 0:
            self[self.lbl_widget].set_no_show_all(True)
            self[self.sep_widget].set_no_show_all(True)
            self[self.lbl_widget].hide()
            self[self.sep_widget].hide()
        else:
            self._add_child_view(self.i_table, self.i_box)

    def update_matrices(self, model):
        for i, (inp, check) in enumerate(zip(self.i_inputs, self.i_checks)):
            prop = self.props[i]
            inp.set_value(getattr(model, prop.label))
            if prop.inherit_flag is not None:
                # Set checkbox sensitivity:
                inh_from = rec_getattr(model, prop.inherit_flag, None)
                check.set_sensitive(not inh_from is None)
                # Set checkbox state:
                inh_value = getattr(model, prop.inherit_flag)
                check.set_active(inh_value)
                # Set inherit value sensitivity
                inp.set_sensitive(not inh_value)
            elif check is not None:
                check.set_senstive(False)

    pass # end of class

class MatrixView(HasChildView, ProbabilityViewMixin, BaseView):
    """
        Generic view that is able to generate and update a P and W 'matrix'
        table with labels having correct tooltips (e.g. P110). Can be used for
        any combination of R, G and rank.
    """
    builder = resource_filename(__name__, "glade/matrix.glade")
    top = "base_matrix_table"

    def __init__(self, R, G, rank, **kwargs):
        """
            Eventhough only two of R,G and rank are required theoretically, 
            they are still required by the __init__ function as a validity
            check.
        """
        BaseView.__init__(self, **kwargs)

        # make sure valid params are passed:
        assert(rank == (G ** max(R, 1)))
        self.create_matrices(R, G, rank)

    def create_matrices(self, R, G, rank):
        # calculate moduli for parameter index calculation:
        lR = max(R, 1)
        mod = [0] * lR
        for i in range(lR):
            mod[i] = rank / (G ** (i + 1))
        title_indeces = "".join([chr(105 + i) for i in range(lR + 1)])


        # Generic function for both the W and P matrix labels setup
        def create_labels(rank, table, current_lR, fmt, tooltip=lambda x, y, current_lR, fmt: ""):
            labels = [[None] * rank for _ in range(rank)]
            for x in range(rank):
                for y in range(rank):
                    new_lbl = Gtk.Label(label="")
                    new_lbl.set_tooltip_markup(tooltip(x, y, current_lR, fmt))
                    new_lbl.set_justify(Gtk.Justification.CENTER)
                    table.attach(new_lbl, y, y + 1, x, x + 1, xpadding=5, ypadding=5)
                    labels[x][y] = new_lbl
                    del new_lbl
            return labels

        # Generic functions for the tooltips:
        def diagonal_tooltips(x, y, current_lR, fmt):
            if x == y:
                indeces = [0] * current_lR
                for i in range(current_lR):
                    indeces[i] = (int(x / mod[i + (lR - current_lR)]) % G) + 1
                return fmt % tuple(indeces)
            else:
                return "-"

        def subdiagonal_tooltips(x, y, current_lR, fmt):
            rowsuf = [0] * current_lR # e.g. i,j,k
            colsuf = [0] * current_lR # e.g. l,m,n
            for i in range(current_lR):
                rowsuf[i] = (int(x / mod[i + (lR - current_lR)]) % G) + 1
                colsuf[i] = (int(y / mod[i + (lR - current_lR)]) % G) + 1

            # check if last n-1 and first n-1 of the suffices equal each other:
            visible = True
            for i in range(current_lR - 1):
                if rowsuf[i + 1] != colsuf[i]: visible = False
            if visible:
                return fmt % (tuple(rowsuf) + (colsuf[-1],))
            else:
                return "-"

        # Create the matrices:

        self.w_tables = []
        self.w_labels = []
        self.w_titles = []
        self.w_valids = []

        self.p_tables = []
        self.p_labels = []
        self.p_titles = []
        self.p_valids = []

        def setup_everything(tables, titles, valids, labels, title, rank, current_lR, lbl_fmt, tooltips):
            w_table = Gtk.Table(rank, rank, True)
            tables.append(w_table)
            titles.append(title)
            valids.append("")
            labels.append(create_labels(
                rank, w_table,
                current_lR, lbl_fmt,
                tooltips
            ))

        for current_lR in range(1, lR + 1):
            rank = G ** current_lR

            setup_everything(
                self.w_tables, self.w_titles, self.w_valids, self.w_labels,
                "<b>W<sub>" + title_indeces[:current_lR] + "</sub></b>",
                rank, current_lR,
                "W<sub>" + "%d"*current_lR + "</sub>",
                diagonal_tooltips
            )

            setup_everything(
                self.p_tables, self.p_titles, self.p_valids, self.p_labels,
                "<b>P<sub>" + title_indeces[:current_lR + 1] + "</sub></b>",
                rank, current_lR,
                "P<sub>" + "%d"*(current_lR + 1) + "</sub>",
                subdiagonal_tooltips
            )

        # Add one extra W matrix:
        setup_everything(
            self.w_tables, self.w_titles, self.w_valids, self.w_labels,
            "<b>W<sub>" + title_indeces + "</sub></b>",
            G ** lR, lR,
            "W<sub>" + "%d"*(lR + 1) + "</sub>",
            subdiagonal_tooltips
        )

        self.show_w_matrix(len(self.w_tables) - 2)
        self.show_p_matrix(len(self.w_tables) - 1)

        return

    def update_matrices(self, model):
        lW, lP = model.get_all_matrices()
        def update_matrix(matrix, labels, mask=None, valid=False):
            shape = matrix.shape
            for i in range(shape[0]):
                for j in range(shape[1]):
                    markup = "<small><span foreground=\"%s\">%.3f</span></small>"
                    if mask is not None:
                        fgcol = "#AA0000" if mask[i, j] < 1 else "#00AA00"
                    else:
                        fgcol = "#000000"
                    labels[i][j].set_markup(markup % (fgcol, matrix[i, j]))

        for i, W in enumerate(lW):
            update_matrix(W, self.w_labels[i], model.W_valid_mask[i])
            fgcol, msg = ("#00AA00", "valid") if model.W_valid[i] else ("#AA0000", "invalid")
            self.w_valids[i] = "<small><span foreground=\"%s\">%s</span></small>" % (fgcol, msg)
            self["lbl_W_valid"].set_markup(self.w_valids[i])
        for i, P in enumerate(lP):
            update_matrix(P, self.p_labels[i], model.P_valid_mask[i])
            fgcol, msg = ("#00AA00", "valid") if model.P_valid[i] else ("#AA0000", "invalid")
            self.p_valids[i] = "<small><span foreground=\"%s\">%s</span></small>" % (fgcol, msg)
            self["lbl_P_valid"].set_markup(self.p_valids[i])

    def show_w_matrix(self, index):
        index = max(min(index, len(self.w_tables) - 1), 0)
        self._add_child_view(self.w_tables[index], self['w_box'])
        self["lbl_W_title"].set_markup(self.w_titles[index])
        self["lbl_W_valid"].set_markup(self.w_valids[index])
        self.show_all()
        return index

    def show_p_matrix(self, index):
        index = max(min(index, len(self.p_tables) - 1), 0)
        self._add_child_view(self.p_tables[index], self['p_box'])
        self["lbl_P_title"].set_markup(self.p_titles[index])
        self["lbl_P_valid"].set_markup(self.w_valids[index])
        self.show_all()
        return index

    pass # end of class




# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import os, csv

import numpy as np

from pyxrd.generic.io.utils import get_case_insensitive_glob
from pyxrd.generic.utils import u

from ..csv_base_parser import CSVBaseParser
from .namespace import xrd_parsers
from .xrd_parser_mixin import XRDParserMixin

class GenericXYCSVParser(XRDParserMixin, CSVBaseParser):
    """
        Generic xy-data CSV parser. Does not care about extensions. 
        Should be sub-classed!
    """

    description = "ASCII XY data"

    default_fmt_params = {
        "delimiter": ',',
        "doublequote": True,
        "escapechar": None,
        "quotechar": "\"",
        "quoting": csv.QUOTE_MINIMAL,
        "skipinitialspace": True,
        "strict": False
    }

    @classmethod
    def _parse_header(cls, filename, fp, data_objects=None, close=False, split_columns=True, has_header=True, file_start=0, **fmt_params):
        fmt_params = dict(cls.default_fmt_params, **fmt_params)
        f = fp

        # Goto start of file
        f.seek(file_start)

        # Get base filename:
        try:
            basename = u(os.path.basename(filename))
        except:
            basename = None

        # Read in the first and last data line and put the file cursor back
        # at its original position
        header = f.readline().strip()
        if not has_header:
            f.seek(file_start) # go back to the start, we still use the first line as header, but also as data
        data_start_pos = f.tell()
        first_line = f.readline().strip()
        twotheta_count, last_line = cls.get_last_line(f)
        last_line = last_line.strip()
        f.seek(data_start_pos)

        # Extract the data from the first & last data lines:
        first_line_vals = cls.parse_raw_line(first_line, float, **fmt_params)
        last_line_vals = cls.parse_raw_line(last_line, float, **fmt_params)
        num_samples = len(first_line_vals) - 1 # first column is 2-theta
        twotheta_min = first_line_vals[0]
        twotheta_max = last_line_vals[0]
        twotheta_step = int((twotheta_max - twotheta_min) / twotheta_count)

        # Parse the header line:
        sample_names = cls.parse_raw_line(header, lambda s: s, **fmt_params)[1:]
        if len(sample_names) < num_samples:
            sample_names.extend([""] * (num_samples - len(sample_names)))
        if len(sample_names) > num_samples:
            sample_names = sample_names[:num_samples]

        # Adapt DataObject list
        data_objects = cls._adapt_data_object_list(data_objects, num_samples=num_samples)

        # Fill in header info:
        for i, sample_name in enumerate(sample_names):
            data_objects[i].update(
                filename=basename,
                name=sample_name,
                twotheta_min=twotheta_min,
                twotheta_max=twotheta_max,
                twotheta_step=twotheta_step,
                twotheta_count=twotheta_count
            )

        if close: f.close()
        return data_objects
    @classmethod
    def _parse_data(cls, filename, fp, data_objects=None, close=False, split_columns=True, has_header=True, **fmt_params):
        f = fp

        if f is not None:
            for row in csv.reader(f, **fmt_params):
                if row:
                    data = list(map(float, row))
                    x, ay = data[0], data[1:] # ay contains columns with y values
                    for data_object, y in zip(data_objects, ay):
                        if getattr(data_object, "data", None) is None:
                            data_object.data = []
                        data_object.data.append([x, y])

            for data_object in data_objects:
                data_object.data = np.array(data_object.data)

        if close: f.close()
        return data_objects

    @classmethod
    def parse(cls, fp, data_objects=None, close=True, split_columns=True, **fmt_params):
        """
            Files are sniffed for the used csv dialect, but an optional set of
            formatting parameters can be passed that will override the sniffed
            parameters.
        """
        filename, f, close = cls._get_file(fp, close=close)

        # Guess dialect:
        fmt_params, has_header, file_start = cls.sniff(f, **fmt_params)

        # Parse header:
        data_objects = cls._parse_header(filename, f, data_objects=data_objects, split_columns=split_columns, has_header=has_header, file_start=file_start, **fmt_params)

        # Parse data:
        data_objects = cls._parse_data(filename, f, data_objects=data_objects, split_columns=split_columns, has_header=has_header, **fmt_params)

        if close: f.close()
        return data_objects

    pass # end of class

@xrd_parsers.register_parser()
class CSVParser(GenericXYCSVParser):
    """
        ASCII *.DAT, *.CSV, *.TAB and *.XY format parser
    """

    description = "CSV XRD data"
    extensions = get_case_insensitive_glob("*.DAT", "*.CSV", "*.TAB", "*.XY")

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

from .abstract_adapter import AbstractAdapter

class DummyAdapter(AbstractAdapter):
    """
        An adapter that does nothing. Really nothing.
    """

    def __init__(self, controller=None, prop=None, widget=None):
        super(DummyAdapter, self).__init__(controller, prop, widget)

    def _connect_widget(self):
        pass # nothing to do

    def _disconnect_widget(self, widget=None):
        pass # nothing to do

    def _connect_model(self):
        pass # nothing to do

    def _disconnect_model(self, model=None):
        pass # nothing to do

    def _read_widget(self):
        pass # nothing to do

    def _write_widget(self, val):
        pass # nothing to do

    def _read_property(self, *args):
        pass # nothing to do

    def _write_property(self, value, *args):
        pass # nothing to do

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk

sel_modes = {
    'NONE': Gtk.SelectionMode.NONE,
    'SINGLE': Gtk.SelectionMode.SINGLE,
    'BROWSE': Gtk.SelectionMode.BROWSE,
    'MULTIPLE': Gtk.SelectionMode.MULTIPLE
}

sort_types = {
    'ASCENDING': Gtk.SortType.ASCENDING,
    'DESCENDING': Gtk.SortType.DESCENDING
}

from .cell_renderer_tools import get_default_renderer, parse_callback, parse_kwargs

class PyXRDTreeViewColumn(Gtk.TreeViewColumn):
    """
        A custom TreeViewColumn that stores information about its attribute 
        mappings and provides acces to them with the get_col_attr function.
    """

    def __init__(self, title=None, cell_renderer=None, **kwargs):
        Gtk.TreeViewColumn.__init__(self, title, cell_renderer)
        self._attrs = dict()
        self.set_attributes(cell_renderer, **kwargs)

    def set_attributes(self, cell_renderer, **kwargs):
        for key, val in kwargs.items():
            self._attrs[key] = val
        Gtk.TreeViewColumn.set_attributes(self, cell_renderer, **kwargs)

    def add_attribute(self, cell_renderer, attribute, column):
        self._attrs[attribute] = column
        Gtk.TreeViewColumn.set_attributes(self, cell_renderer, attribute, column)

    def clear_attributes(self, cell_renderer):
        self._attrs = dict()
        Gtk.TreeViewColumn.clear_attributes(self, cell_renderer)

    def get_col_attr(self, attr):
        return self._attrs.get(attr, -1)

def _get_default_column(title, rend,
        data_func=None,
        spacing=0,
        visible=True,
        resizable=True,
        sizing=1,
        fixed_width=-1,
        min_width=-1,
        max_width=-1,
        expand=True,
        clickable=False,
        alignment=0.0,
        reorderable=False,
        sort_column_id=-1,
        sort_indicator=False,
        sort_order='ASCENDING',
        col_attrs={}):
    """
        Creates a PyXRDTreeViewColumn using the arguments passed. Column 
        attribute mappings are to be passed as a single dict,
        not as key-word arguments.
    """
    try:
        sort_order = sort_types[sort_order]
    except KeyError as err:
        raise ValueError("Invalid value '%s' for sort order!" % sort_order) from err 
    col = PyXRDTreeViewColumn(title, rend, **col_attrs)
    if data_func is not None:
        callback, args = parse_callback(data_func)
        col.set_cell_data_func(rend, callback, args)
    col.set_spacing(spacing)
    col.set_visible(visible)
    col.set_resizable(resizable)
    col.set_sizing(sizing)
    if fixed_width >= 0:
        col.set_sizing(Gtk.TreeViewColumnSizing.Fixed)
        col.set_fixed_width(fixed_width)
    else:
        col.set_sizing(Gtk.TreeViewColumnSizing.GROW_ONLY)
    col.set_min_width(min_width)
    col.set_max_width(max_width)
    col.set_title(title)
    col.set_expand(expand)
    col.set_clickable(clickable)
    col.set_alignment(alignment)
    col.set_reorderable(reorderable)
    col.set_sort_column_id(sort_column_id)
    col.set_sort_indicator(sort_indicator)
    col.set_sort_order(sort_order)
    col.set_resizable(resizable)
    col.set_expand(expand)
    col.set_alignment(alignment)
    return col

def new_text_column(title,
        edited_callback=None,
        data_func=None,
        spacing=0,
        visible=True,
        resizable=True,
        sizing=1,
        fixed_width=-1,
        min_width=-1,
        max_width=-1,
        expand=True,
        clickable=False,
        alignment=None,
        reorderable=False,
        sort_column_id=-1,
        sort_indicator=False,
        sort_order='ASCENDING',
        **kwargs):
    """
        Creates a TreeViewColumn packed with a CellRendererText .
    """
    kwargs, col_attrs = parse_kwargs(**kwargs)
    alignment = alignment if alignment is not None else kwargs["xalign"]

    rend = get_default_renderer(Gtk.CellRendererText, **kwargs)
    if edited_callback is not None:
        callback, args = parse_callback(edited_callback, reduce=False)
        rend.connect('edited', callback, *args)

    col = _get_default_column(
        title, rend,
        data_func=data_func,
        spacing=spacing,
        visible=visible,
        resizable=resizable,
        sizing=sizing,
        fixed_width=fixed_width,
        min_width=min_width,
        max_width=max_width,
        expand=expand,
        clickable=clickable,
        alignment=alignment,
        reorderable=reorderable,
        sort_column_id=sort_column_id,
        sort_indicator=sort_indicator,
        sort_order=sort_order,
        col_attrs=col_attrs)
    return col

def new_pb_column(title,
        data_func=None,
        spacing=0,
        visible=True,
        resizable=True,
        sizing=1,
        fixed_width=-1,
        min_width=-1,
        max_width=-1,
        expand=True,
        clickable=False,
        alignment=None,
        reorderable=False,
        sort_column_id=-1,
        sort_indicator=False,
        sort_order='ASCENDING',
        **kwargs):
    """
        Creates a TreeViewColumn packed with a CellRendererPixbuf.
    """
    kwargs, col_attrs = parse_kwargs(**kwargs)
    alignment = alignment if alignment is not None else kwargs["xalign"]

    rend = get_default_renderer(Gtk.CellRendererPixbuf, **kwargs)

    col = _get_default_column(
        title, rend,
        data_func=data_func,
        spacing=spacing,
        visible=visible,
        resizable=resizable,
        sizing=sizing,
        fixed_width=fixed_width,
        min_width=min_width,
        max_width=max_width,
        expand=expand,
        clickable=clickable,
        alignment=alignment,
        reorderable=reorderable,
        sort_column_id=sort_column_id,
        sort_indicator=sort_indicator,
        sort_order=sort_order,
        col_attrs=col_attrs)
    return col

def new_toggle_column(title,
        data_func=None,
        toggled_callback=None,
        spacing=0,
        visible=True,
        resizable=True,
        sizing=1,
        fixed_width=-1,
        min_width=-1,
        max_width=-1,
        expand=True,
        clickable=False,
        alignment=None,
        reorderable=False,
        sort_column_id=-1,
        sort_indicator=False,
        sort_order='ASCENDING',
        **kwargs):
    """
        Creates a TreeViewColumn packed with a CellRendererToggle.
    """
    kwargs, col_attrs = parse_kwargs(**kwargs)
    alignment = alignment if alignment is not None else kwargs["xalign"]

    rend = get_default_renderer(Gtk.CellRendererToggle, **kwargs)
    if toggled_callback is not None:
        callback, args = parse_callback(toggled_callback, reduce=False)
        rend.connect('toggled', callback, *args)

    col = _get_default_column(
        title, rend,
        data_func=data_func,
        spacing=spacing,
        visible=visible,
        resizable=resizable,
        sizing=sizing,
        fixed_width=fixed_width,
        min_width=min_width,
        max_width=max_width,
        expand=expand,
        clickable=clickable,
        alignment=alignment,
        reorderable=reorderable,
        sort_column_id=sort_column_id,
        sort_indicator=sort_indicator,
        sort_order=sort_order,
        col_attrs=col_attrs)
    return col

def new_combo_column(title,
        data_func=None,
        changed_callback=None,
        edited_callback=None,
        editing_started_callback=None,
        editing_canceled_callback=None,
        spacing=0,
        visible=True,
        resizable=True,
        sizing=1,
        fixed_width=-1,
        min_width=-1,
        max_width=-1,
        expand=True,
        clickable=False,
        alignment=None,
        reorderable=False,
        sort_column_id=-1,
        sort_indicator=False,
        sort_order='ASCENDING',
        **kwargs):
    """
        Creates a TreeViewColumn packed with a CellRendererCombo.
    """
    kwargs, col_attrs = parse_kwargs(**kwargs)
    alignment = alignment if alignment is not None else kwargs["xalign"]

    rend = get_default_renderer(Gtk.CellRendererCombo, **kwargs)
    if changed_callback is not None:
        callback, args = parse_callback(changed_callback, reduce=False)
        rend.connect('changed', callback, *args)
    if edited_callback is not None:
        callback, args = parse_callback(edited_callback, reduce=False)
        rend.connect('edited', callback, *args)
    if editing_started_callback is not None:
        callback, args = parse_callback(editing_started_callback, reduce=False)
        rend.connect('editing-started', callback, *args)
    if editing_canceled_callback is not None:
        callback, args = parse_callback(editing_canceled_callback, reduce=False)
        rend.connect('editing-canceled', callback, *args)

    col = _get_default_column(
        title, rend,
        data_func=data_func,
        spacing=spacing,
        visible=visible,
        resizable=resizable,
        sizing=sizing,
        fixed_width=fixed_width,
        min_width=min_width,
        max_width=max_width,
        expand=expand,
        clickable=clickable,
        alignment=alignment,
        reorderable=reorderable,
        sort_column_id=sort_column_id,
        sort_indicator=sort_indicator,
        sort_order=sort_order,
        col_attrs=col_attrs)

    return col

def create_float_data_func(attribute='text', fmt="%.5f", invalid="#NA#"):
    """
        Creates a data function that can be used to render floats as formatted
        strings, with detection of invalid values (e.g. None)
    """
    def float_renderer(column, cell, model, itr, args=None):
        nr = model.get_value(itr, column.get_col_attr(attribute))
        try:
            cell.set_property('text', fmt % nr)
        except:
            cell.set_property('text', invalid)
    return float_renderer

def reset_columns(tv):
    """
        Remove all columns from the treeview
    """
    for col in tv.get_columns():
        tv.remove_column(col)

def setup_treeview(tv, model,
        reset=False,
        on_cursor_changed=None,
        on_selection_changed=None,
        sel_mode='SINGLE'):
    """
        Sets up a treeview (signal connection, sets selection mode).
    """
    try:
        sel_mode = sel_modes[sel_mode]
    except KeyError as err:
        raise ValueError("Invalid value '%s' for selection mode!" % sel_mode) from err
    if reset: reset_columns(tv)
    sel = tv.get_selection()
    sel.set_mode(sel_mode)
    ids = ()
    if on_cursor_changed is not None:
        ids += (tv.connect('cursor_changed', on_cursor_changed),)
    if on_selection_changed is not None:
        ids += (sel.connect('changed', on_selection_changed),)
    return ids




# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

import numpy as np
from scipy import stats
from scipy.interpolate import interp1d

from mvc.models.properties import (
    StringProperty, BoolProperty, FloatProperty,
    StringChoiceProperty, SignalMixin, ListProperty
)

from pyxrd.data import settings

from pyxrd.generic.utils import not_none
from pyxrd.generic.io.custom_io import storables
from pyxrd.generic.models.properties import InheritableMixin
from pyxrd.calculations.peak_detection import multi_peakdetect

from .storable_xy_data import StorableXYData

#from pyxrd.file_parsers.ascii_parser import ASCIIParser

@storables.register()
class PyXRDLine(StorableXYData):
    """
        A PyXRDLine is an abstract attribute holder for a real 'Line' object,
        whatever the plotting library used may be. Attributes are line width and
        color.        
    """

    # MODEL INTEL:
    class Meta(StorableXYData.Meta):
        store_id = "PyXRDLine"

    # OBSERVABLE PROPERTIES:

    #: The line label
    label = StringProperty(
        default="", text="Label", persistent=True
    )

    #: The line color
    color = StringProperty(
        default="#000000", text="Label",
        visible=True, persistent=True, widget_type="color",
        inherit_flag="inherit_color", inherit_from="parent.parent.display_exp_color",
        signal_name="visuals_changed",
        mix_with=(InheritableMixin, SignalMixin)
    )

    #: Flag indicating whether to use the grandparents color yes/no
    inherit_color = BoolProperty(
        default=True, text="Inherit color",
        visible=True, persistent=True,
        signal_name="visuals_changed", mix_with=(SignalMixin,)
    )

    #: The linewidth in points
    lw = FloatProperty(
        default=2.0, text="Linewidth",
        visible=True, persistent=True, widget_type="spin",
        inherit_flag="inherit_lw", inherit_from="parent.parent.display_exp_lw",
        signal_name="visuals_changed",
        mix_with=(InheritableMixin, SignalMixin),
    )

    #: Flag indicating whether to use the grandparents linewidth yes/no
    inherit_lw = BoolProperty(
        default=True, text="Inherit linewidth",
        visible=True, persistent=True,
        signal_name="visuals_changed", mix_with=(SignalMixin,),
    )

    #: A short string describing the (matplotlib) linestyle
    ls = StringChoiceProperty(
        default=settings.EXPERIMENTAL_LINESTYLE, text="Linestyle",
        visible=True, persistent=True, choices=settings.PATTERN_LINE_STYLES,
        mix_with=(InheritableMixin, SignalMixin,), signal_name="visuals_changed",
        inherit_flag="inherit_ls", inherit_from="parent.parent.display_exp_ls",
    )

    #: Flag indicating whether to use the grandparents linestyle yes/no
    inherit_ls = BoolProperty(
        default=True, text="Inherit linestyle",
        visible=True, persistent=True,
        mix_with=(SignalMixin,), signal_name="visuals_changed"
    )

    #: A short string describing the (matplotlib) marker
    marker = StringChoiceProperty(
        default=settings.EXPERIMENTAL_MARKER, text="Marker",
        visible=True, persistent=True, choices=settings.PATTERN_MARKERS,
        mix_with=(InheritableMixin, SignalMixin,), signal_name="visuals_changed",
        inherit_flag="inherit_marker", inherit_from="parent.parent.display_exp_marker",
    )

    #: Flag indicating whether to use the grandparents linewidth yes/no
    inherit_marker = BoolProperty(
        default=True, text="Inherit marker",
        visible=True, persistent=True,
        mix_with=(SignalMixin,), signal_name="visuals_changed",
    )
    
    #: z-data (e.g. relative humidity, temperature, for multi-column 'lines')
    z_data = ListProperty(
        default=None, text="Z data", data_type=float,
        persistent=True, visible=False
    ) 

    # REGULAR PROPERTIES:   
    @property
    def max_display_y(self):
        if self.num_columns > 2:
            # If there's several y-columns, check if we have z-data associated with them
            # if so, it is a 2D pattern, otherwise this is a multi-line pattern
            if len(self.z_data) > 2:
                return np.max(self.z_data)
            else:
                return self.max_y
        else:
            # If there's a single comumn of y-data, just get the max value
            return self.max_y

    @property
    def min_intensity(self):
        if self.num_columns > 2:
            return np.min(self.z_data)
        else:        
            return self.min_y

    @property
    def abs_max_intensity(self):
        return self.abs_max_y

    # ------------------------------------------------------------
    #      Initialisation and other internals
    # ------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        """
            Valid keyword arguments for a PyXRDLine are:
                data: the actual data containing x and y values
                label: the label for this line
                color: the color of this line
                inherit_color: whether to use the parent-level color or its own
                lw: the line width of this line
                inherit_lw: whether to use the parent-level line width or its own
                ls: the line style of this line
                inherit_ls: whether to use the parent-level line style or its own
                marker: the line marker of this line
                inherit_marker: whether to use the parent-level line marker or its own
                z_data: the z-data associated with the columns in a multi-column pattern
        """
        my_kwargs = self.pop_kwargs(kwargs,
            *[prop.label for prop in PyXRDLine.Meta.get_local_persistent_properties()]
        )
        super(PyXRDLine, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        with self.visuals_changed.hold():
            self.label = self.get_kwarg(kwargs, self.label, "label")
            self.color = self.get_kwarg(kwargs, self.color, "color")
            self.inherit_color = bool(self.get_kwarg(kwargs, self.inherit_color, "inherit_color"))
            self.lw = float(self.get_kwarg(kwargs, self.lw, "lw"))
            self.inherit_lw = bool(self.get_kwarg(kwargs, self.inherit_lw, "inherit_lw"))
            self.ls = self.get_kwarg(kwargs, self.ls, "ls")
            self.inherit_ls = bool(self.get_kwarg(kwargs, self.inherit_ls, "inherit_ls"))
            self.marker = self.get_kwarg(kwargs, self.marker, "marker")
            self.inherit_marker = bool(self.get_kwarg(kwargs, self.inherit_marker, "inherit_marker"))
            self.z_data = list(self.get_kwarg(kwargs, [0], "z_data"))

    # ------------------------------------------------------------
    #      Input/Output stuff
    # ------------------------------------------------------------
    @classmethod
    def from_json(cls, **kwargs): # @ReservedAssignment
        if "xy_store" in kwargs:
            if "type" in kwargs["xy_store"]:
                kwargs["data"] = kwargs["xy_store"]["properties"]["data"]
        elif "xy_data" in kwargs:
            if "type" in kwargs["xy_data"]:
                kwargs["data"] = kwargs["xy_data"]["properties"]["data"]
            kwargs["label"] = kwargs["data_label"]
            del kwargs["data_name"]
            del kwargs["data_label"]
            del kwargs["xy_data"]
        return cls(**kwargs)

    # ------------------------------------------------------------
    #      Convenience Methods & Functions
    # ------------------------------------------------------------
    def interpolate(self, *x_vals, **kwargs):
        """
            Returns a list of (x, y) tuples for the passed x values. An optional
            column keyword argument can be passed to select a column, by default
            the first y-column is used. Returned y-values are interpolated. 
        """
        column = kwargs.get("column", 0)
        f = interp1d(self.data_x, self.data_y[:, column])
        return list(zip(x_vals, f(x_vals)))

    def get_plotted_y_at_x(self, x):
        """
            Gets the (interpolated) plotted value at the given x position.
            If this line has not been plotted (or does not have
            access to a '__plot_line' attribute set by the plotting routines)
            it will return 0.
        """
        try:
            xdata, ydata = getattr(self, "__plot_line").get_data()
        except AttributeError:
            logging.exception("Attribute error when trying to get plotter data at x position!")
        else:
            if len(xdata) > 0 and len(ydata) > 0:
                return np.interp(x, xdata, ydata)
        return 0

    def calculate_npeaks_for(self, max_threshold, steps):
        """
            Calculates the number of peaks for `steps` threshold values between
            0 and `max_threshold`. Returns a tuple containing two lists with the
            threshold values and the corresponding number of peaks. 
        """
        length = self.data_x.size

        resolution = length / (self.data_x[-1] - self.data_x[0])
        delta_angle = 0.05
        window = int(delta_angle * resolution)
        window += (window % 2) * 2

        steps = max(steps, 2) - 1
        factor = max_threshold / steps

        deltas = [i * factor for i in range(0, steps)]

        numpeaks = []

        maxtabs, mintabs = multi_peakdetect(self.data_y[:, 0], self.data_x, 5, deltas)
        for maxtab, _ in zip(maxtabs, mintabs):
            numpeak = len(maxtab)
            numpeaks.append(numpeak)
        numpeaks = list(map(float, numpeaks))

        return deltas, numpeaks

    def get_best_threshold(self, max_threshold=None, steps=None, status_dict=None):
        """
            Estimates the best threshold for peak detection using an
            iterative algorithm. Assumes there is a linear contribution from noise.
            Returns a 4-tuple containing the selected threshold, the maximum
            threshold, a list of threshold values and a list with the corresponding
            number of peaks.
        """
        length = self.data_x.size
        steps = not_none(steps, 20)
        threshold = 0.1
        max_threshold = not_none(max_threshold, threshold * 3.2)

        def get_new_threshold(threshold, deltas, num_peaks, ln):
            # Left side line:
            x = deltas[:ln]
            y = num_peaks[:ln]
            slope, intercept, R, _, _ = stats.linregress(x, y)
            return R, -intercept / slope

        if length > 2:
            # Adjust the first distribution:
            deltas, num_peaks = self.calculate_npeaks_for(max_threshold, steps)

            #  Fit several lines with increasing number of points from the
            #  generated threshold / marker count graph. Stop when the
            #  R-coefficiënt drops below 0.95 (past linear increase from noise)
            #  Then repeat this by increasing the resolution of data points
            #  and continue until the result does not change anymore

            last_threshold = None
            solution = False
            max_iters = 10
            min_iters = 3
            itercount = 0
            if status_dict is not None:
                status_dict["progress"] = 0

            while not solution:
                # Number of points to use for the lin regress:
                ln = 4
                # Maximum number of points to use:
                max_ln = len(deltas)
                # Flag indicating if we can stop searching for the linear part
                stop = False
                while not stop:
                    R, threshold = get_new_threshold(threshold, deltas, num_peaks, ln)
                    max_threshold = threshold * 3.2
                    if abs(R) < 0.98 or ln >= max_ln:
                        stop = True
                    else:
                        ln += 1
                itercount += 1 # Increase # of iterations
                if last_threshold:
                    # Check if we have run at least `min_iters`, at most `max_iters`
                    # and have not reached an equilibrium.
                    solution = bool(
                        itercount > min_iters and not
                        (
                            itercount <= max_iters and
                            last_threshold - threshold >= 0.001
                        )
                    )
                    if not solution:
                        deltas, num_peaks = self.calculate_npeaks_for(max_threshold, steps)
                last_threshold = threshold
                if status_dict is not None:
                    status_dict["progress"] = float(itercount / max_iters)

            return (deltas, num_peaks), threshold, max_threshold
        else:
            return ([], []), threshold, max_threshold


    pass # end of class


#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import unittest

def run_all_tests(*args, **kwargs):
    all_tests = get_all_tests()
    unittest.TextTestRunner().run(all_tests)

def get_all_tests():
    return unittest.TestLoader().discover('.')



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

class MetaParser(type):
    """
        Metatype for the parser sub classes, allowing for auto file filter
        creation.
    """
    def __new__(meta, name, bases, attrs): # @NoSelf
        res = super(MetaParser, meta).__new__(meta, name, bases, attrs)
        res.setup_file_filter()
        return res


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from mvc.models.properties.observe_mixin import ObserveMixin

from mvc.support.utils import rec_getattr

class ObserveChildMixin(ObserveMixin):
    """
    A descriptor mixin that will make the instance observe and relieve the
    objects set and clear and set the parent property on the old and new object respectively
    """

    def __relieve_old(self, instance, old, new):
        if old is not None:
            instance.relieve_model(old)
            old.parent = None

    def __observe_new(self, instance, old, new):
        if new is not None:
            new.parent = instance
            instance.observe_model(new)

    pass

class InheritableMixin(object):
    """
    Mixing for the ~:class:`mvc.models.properties.LabeledProperty` descriptor
    that allows the property to be inheritable from another property.
    When this Mixin is used, the user should pass two additional keyword 
    arguments to the descriptor:
        - inheritable: boolean set to True if inheriting should be enabled
        - inherit_flag: dotted string describing where to get the flag 
          indicating the property is inherited yes/no
        - inherit_from: dotted string describing where to get the attribute if
          the inherit_flag is True 
    """

    inheritable = True
    inherit_flag = None
    inherit_from = None

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        value = self.get_uninherited(instance, owner)
        if self.inheritable and rec_getattr(instance, self.inherit_flag, False):
            value = rec_getattr(instance, self.inherit_from, value)
        return value

    def get_uninherited(self, instance, owner=None):
        return super(InheritableMixin, self).__get__(instance, owner)

    pass #end of class

class IndexProperty(object):
    """Descriptor used to create indexable properties (e.g. W[1,1])"""

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        if doc is None and fget is not None and hasattr(fget, "__doc__"):
            doc = fget.__doc__
        self._get = fget
        self._set = fset
        self._del = fdel
        self.__doc__ = doc

    def __get__(self, instance, owner):
        if instance is None:
            return self
        else:
            return BoundIndexProperty(self, instance)

    def __set__(self, instance, value):
        raise AttributeError("can't set attribute")

    def __delete__(self, instance):
        raise AttributeError("can't delete attribute")

    def getter(self, fget):
        return IndexProperty(fget, self._set, self._del, self.__doc__)

    def setter(self, fset):
        return IndexProperty(self._get, fset, self._del, self.__doc__)

    def deleter(self, fdel):
        return IndexProperty(self._get, self._set, fdel, self.__doc__)

    pass #end of class

class BoundIndexProperty(object):

    def __init__(self, item_property, instance):
        self.__item_property = item_property
        self.__instance = instance

    def __getitem__(self, key):
        fget = self.__item_property._get
        if fget is None:
            raise AttributeError("unreadable attribute item")
        return fget(self.__instance, key)

    def __setitem__(self, key, value):
        fset = self.__item_property._set
        if fset is None:
            raise AttributeError("can't set attribute item")
        fset(self.__instance, key, value)

    def __delitem__(self, key):
        fdel = self.__item_property._del
        if fdel is None:
            raise AttributeError("can't delete attribute item")
        fdel(self.__instance, key)

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from pyxrd.generic.controllers.dialog_controller import DialogController
from pyxrd.refinement.parspace import ParameterSpaceGenerator

class RefinerController(DialogController):
    """
        A controller for a Refiner object that keeps track
        of the solutions and residuals generated by the refinement
        algorithm. This allows to show a nice dialog with the end
        results and some graphs about the parameter space.
    """

    auto_adapt_excluded = [
        "refine_method_index",
        "refinables",
        "make_psp_plots",
    ]

    register_lazy = False
    samples = None

    def __init__(self, refiner, *args, **kwargs):
        super(RefinerController, self).__init__(*args, **kwargs)
        self.refiner = refiner

    # ------------------------------------------------------------
    #      GTK Signal handlers
    # ------------------------------------------------------------
    def on_cancel(self):
        self.view.hide()
        self.parent.parent.view.parent.show()

    def on_btn_initial_clicked(self, event):
        self.refiner.apply_initial_solution()
        del self.refiner
        self.on_cancel()
        return True

    def on_btn_best_clicked(self, event):
        self.refiner.apply_best_solution()
        del self.refiner
        self.on_cancel()
        return True

    def on_btn_last_clicked(self, event):
        self.refiner.apply_last_solution()
        del self.refiner
        self.on_cancel()
        return True

    # ------------------------------------------------------------
    #      Methods & Functions
    # ------------------------------------------------------------
    def update_labels(self):
        self.view.update_labels(
            self.refiner.history.initial_residual,
            self.refiner.history.best_residual,
            self.refiner.history.last_residual,
        )
    
    def generate_images(self, output_dir="", density=200):
        """
            Generate the parameter space plots
        """
        samples = self.refiner.get_plot_samples()
        labels = self.refiner.get_plot_labels()
        truths = self.refiner.history.best_solution
       
        psg = ParameterSpaceGenerator()
        psg.initialize(self.refiner.ranges, 199)
        for sample in samples:
            psg.record(sample[:-1], sample[-1])
        psg.plot_images(self.view.figure, truths, labels[:-1])

    def clear_images(self):
        self.view.figure.clear()
        self.view.figure.text(0.5, 0.5, "no plots to show", va="center", ha="center")

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

from mvc.support.utils import rec_getattr

class SetActionMixin(object):
    """
    A descriptor mixin that will invoke a method on the instance
    owning this property after setting it.
    
    Expects two more keyword arguments to be passed to the property constructor:
        - set_action_name: a dotted string describing where to get the method
          from the instance
        - set_action_before: flag indicating whether this action should be 
          invoked before setting the property (default False) 
    """

    set_action_name = None
    set_action_before = False

    def __set__(self, instance, value):
        action = rec_getattr(instance, self.set_action_name, None)
        assert callable(action), "The action in a SetActionMixin (%s) should be callable!" % self.label
        if self.set_action_before: action()
        super(SetActionMixin, self).__set__(instance, value)
        if not self.set_action_before: action()


    pass # end of class

class GetActionMixin(object):
    """
    A descriptor mixin that will invoke a method on the instance
    owning this property before getting it.
    
    Expects two more keyword arguments to be passed to the property constructor:
        - get_action_name: a dotted string describing where to get the method
          from the instance
        - get_action_after: flag indicating whether this action should be 
          invoked after setting the property (default False) 
    """

    get_action_name = None
    get_action_after = False

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        action = rec_getattr(instance, self.get_action_name, None)
        assert callable(action), "The action in a GetActionMixin (%s) should be callable!" % self.label
        if self.get_action_after: action()
        value = super(GetActionMixin, self).__get__(instance, owner=owner)
        if not self.get_action_after: action()
        return value

    pass # end of class




#!/usr/bin/python

# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import gi
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
from gi.repository import GObject  # @UnresolvedImport

import unittest

from mvc.adapters.gtk_support.treemodels import BaseObjectListStore

__all__ = [
    'TestBaseObjectListStore',
]

class _DummyObject(object):
    class Meta(object):
        @classmethod
        def get_column_properties(cls):
            return [
                ["name", str],
                ["number", float],
                ["test", object]
            ]

    def __init__(self, *args, **kwargs):
        super(_DummyObject, self).__init__()
        for key, val in kwargs.items():
            setattr(self, key, val)

    pass # end of class

class TestBaseObjectListStore(unittest.TestCase):

    def setUp(self):
        self.store = BaseObjectListStore(_DummyObject)

    def tearDown(self):
        del self.store

    def test_setup(self):
        self.assertNotEqual(self.store, None)

    def test_columns(self):
        self.assertEqual(self.store.get_n_columns(), len(self.store._class_type.Meta.get_column_properties()))
        self.assertEqual(self.store.get_column_type(self.store.c_name), GObject.type_from_name("gchararray"))
        self.assertEqual(self.store.get_column_type(self.store.c_number), GObject.type_from_name("gdouble"))
        self.assertEqual(self.store.get_column_type(self.store.c_test), GObject.type_from_name("PyObject"))

    def test_convert(self):
        self.assertEqual(self.store.convert(1, "0.5"), 0.5)

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

"""
    Input/Output related utility functions
"""

import os, sys

from mvc.support.file_utils import get_case_insensitive_glob, retrieve_lowercase_extension, relpath

# Small workaround to provide a unicode-aware open method:
if sys.version_info[0] < 3: # Pre Python 3.0
    import codecs
    _open_func_bak = open # Make a back up, just in case
    open = codecs.open #@ReservedAssignment

def sizeof_fmt(num):
    ''' Returns a human-friendly string when given a size in bytes '''
    for x in ['bytes', 'kB', 'MB', 'GB', 'TB']:
        if num < 1024.0:
            return "%3.1f %s" % (num, x)
        num /= 1024.0

def get_size(path='.', maxsize=None):
    ''' Gets the recursive size of a path, can be limited to a maxsize '''
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(path): #@UnusedVariable
        for f in filenames:
            fp = os.path.join(dirpath, f)
            total_size += os.path.getsize(fp)
            if maxsize is not None and total_size > maxsize:
                break
        if maxsize is not None and total_size > maxsize:
            break
    return total_size

def unicode_open(*args, **kwargs):
    """
        Opens files in UTF-8 encoding by default, unless an 'encoding'
        keyword argument is passed. Returns a file object.
    """
    if not "encoding" in kwargs:
        kwargs["encoding"] = "utf-8"
    return open(*args, **kwargs)


# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

from mvc.support.utils import rec_getattr
from mvc.support import observables

class SignalMixin(object):
    """
    A descriptor mixin that will invoke a signal on the instance
    owning this property when set. 
    
    Expects two more keyword arguments to be passed to the property constructor:
        - signal_name: a dotted string describing where to get the signal object
          from the instance
    """

    signal_name = "data_changed"

    def __set__(self, instance, value):
        signal = rec_getattr(instance, self.signal_name, None)
        if signal is not None:
            # Get the old value
            old = getattr(instance, self.label)
            with signal.ignore():
                super(SignalMixin, self).__set__(instance, value)
            # Get the new value
            new = getattr(instance, self.label)
            # Check if we're dealing with some special class (in case we
            # emit the signal anyway) or immutables (in case we only emit
            # when the value has changed)  
            if isinstance(old, observables.ObsWrapperBase) or old != new:
                signal.emit()
        else:
            super(SignalMixin, self).__set__(instance, value)
            
    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gdk

from .event_delegator import MPLCanvasEventDelegator

class DraggableMixin(object):
    """
        A mixin that can be used to make any matplotlib artist draggable.
        The constructor takes two callbacks, `on_dragged` and `on_released`
        which are called when the artist is dragged and released respectively.
        
        These callbacks are passed 3 coordinate tuples: the starting point of
        the drag operation, the previous cursor position and the current cursor
        position. All of these are in display coordinates. If these need to be 
        converted you can use the `convert_display_to_*` methods where * should
        be one of figure, axes or data. Aside from these coordinates, the callbacks
        are also passed the last event object.
        
        For more advanced usage the user can also override the 
        `_draggable_on_press`, `_draggable_on_motion` and `_draggable_on_release`
        event handlers, be sure to call the base implementation if so.
    """

    lock = None  # only one can be animated at a time
    _prev_event_data = None
    _draggable_artist = None
    _draggable_figure = None
    _draggable_canvas = None
    _draggable_axes = None
    _on_dragged = None
    _on_released = None

    _first_event_data = None
    _prev_event_data = None

    @property
    def _is_dragging(self):
        return not bool(
            self._draggable_artist is None or \
            self._first_event_data is None or \
            self._prev_event_data is None
        )

    def __init__(self,
            artist=None,
            on_dragged=None, on_released=None,
            *args, **kwargs):
        super(DraggableMixin, self).__init__(*args, **kwargs)
        self.update(artist, on_dragged, on_released)

    def update(self, artist=None, on_dragged=None, on_released=None):
        self._on_dragged = on_dragged
        self._on_released = on_released
        self.set_draggable_artist(artist)

    def disconnect(self):
        self.set_draggable_artist(None)

    def set_draggable_artist(self, artist):
        if self._draggable_artist is not None:
            self._draggable_disconnect()

        self._draggable_artist = artist
        self._draggable_figure = artist.figure if artist is not None else None
        self._draggable_canvas = artist.figure.canvas if artist is not None else None
        self._draggable_axes = artist.axes if artist is not None else None

        if self._draggable_artist is not None:
            self._draggable_connect()

    def _convert_to_inverse_transform(self, x, y, transf):
        inverse = transf.inverted()
        return inverse.transform((x, y))

    def convert_display_to_figure(self, x, y):
        return self._convert_to_inverse_transform(
            x, y, self._draggable_figure.transFigure)

    def convert_display_to_axes(self, x, y):
        return self._convert_to_inverse_transform(
            x, y, self._draggable_axes.transAxes)

    def convert_display_to_data(self, x, y):
        return self._convert_to_inverse_transform(
            x, y, self._draggable_axes.transData)

    # -- EVENT HANDLERS:

    def _draggable_on_press(self, event):
        if self._draggable_figure is None: return False

        contains, attrd = self._draggable_artist.contains(event) # @UnusedVariable
        if not contains: return False

        self._prev_event_data = event.x, event.y
        self._first_event_data = self._prev_event_data

        return True

    def _draggable_on_motion(self, event):
        if not self._is_dragging: return

        x0, y0 = self._first_event_data
        x1, y1 = self._prev_event_data
        x2, y2 = event.x, event.y

        if callable(self._on_dragged):
            self._on_dragged(
                (x0, y0), (x1, y1), (x2, y2),
                event=event
            )

        self._prev_event_data = x2, y2

    def _draggable_on_release(self, event):
        if not self._is_dragging: return

        x0, y0 = self._first_event_data
        x1, y1 = self._prev_event_data
        x2, y2 = event.x, event.y

        if callable(self._on_released):
            self._on_released(
                (x0, y0), (x1, y1), (x2, y2),
                event=event
            )

        self._prev_event_data = None
        self._first_event_data = None

    # -- CONECTION & DISCONNECTION:

    __connected = False
    def _draggable_connect(self):
        if not self.__connected:
            delegator = MPLCanvasEventDelegator.wrap_canvas(self._draggable_canvas)

            delegator.connect('button_press_event', self._draggable_on_press)
            delegator.connect('button_release_event', self._draggable_on_release)
            delegator.connect('motion_notify_event', self._draggable_on_motion)

            self.__connected = True

    def _draggable_disconnect(self):
        if self.__connected:
            delegator = MPLCanvasEventDelegator.wrap_canvas(self._draggable_canvas)

            delegator.disconnect('button_press_event', self._draggable_on_press)
            delegator.disconnect('button_release_event', self._draggable_on_release)
            delegator.disconnect('motion_notify_event', self._draggable_on_motion)

            self.__connected = False

    pass #end of class


class DraggableVLine(DraggableMixin):
    """
        A draggable vertical line with a callback called with the new x position
        of the line after the user released it.
    """
    def __init__(self, line, callback=None, window=None):
        super(DraggableVLine, self).__init__(
            artist=line,
            on_dragged=self._on_dragged,
            on_released=self._on_released)
        self.line = line
        self.line_x0 = None
        self.callback = callback
        self.window = window

    def _check_cursor(self, event):
        if self.window is not None:
            change_cursor, _ = self.line.contains(event)
            if not change_cursor:
                self.window.set_cursor(None)
            else:
                arrows = Gdk.Cursor.new(Gdk.CursorType.SB_H_DOUBLE_ARROW) #@UndefinedVariable
                self.window.set_cursor(arrows)

    def _draggable_on_motion(self, event):
        super(DraggableVLine, self)._draggable_on_motion(event)
        self._check_cursor(event)

    def _on_dragged(self, x0_y0, x1_y1, x2_y2, event):
        (x0, y0) = x0_y0
        (x1, y1) = x1_y1
        (x2, y2) = x2_y2
        self._check_cursor(event)
        if self.line_x0 == None:
            self.line_x0 = self.line.get_xdata()[0]

        diff = self.convert_display_to_data(x2, 0)[0] - self.convert_display_to_data(x0, 0)[0]
        x = max(self.line_x0 + diff, 0)
        self.line.set_xdata((x, x))

        self.line.figure.canvas.draw()

    def _on_released(self, x0_y0, x1_y1, x2_y2, event):
        """(x0, y0) = x0_y0
        (x1, y1) = x1_y1
        (x2, y2) = x2_y2"""
        self._check_cursor(event)
        if callable(self.callback):
            self.callback(self.line.get_xdata()[0])
        self.line_x0 = None

        # redraw the full figure
        self.line.figure.canvas.draw()

    pass #end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

from contextlib import contextmanager

from ..support.utils import not_none
from ..observers import Observer

from ..models import Model
from ..models.properties import LabeledProperty

from .abstract_adapter import AbstractAdapter

class ModelAdapter(Observer, AbstractAdapter):
    """
        Model-side implementation of the _AbstractAdapter interface.
    """

    # Mode-side property handling:
    _prop_read = None
    _prop_write = None
    _value_error = None

    @property
    def property_name(self):
        """Returns the property name the adapter is connected to"""
        return self._prop.label

    _ignoring_notifs = False
    @contextmanager
    def _ignore_notifications(self):
        """Context manager to temporarily (and exception-safe) ignore
        property changed notifications (e.g. when we're setting the model
        from the widget and vice versa)"""
        self._ignoring_notifs = True
        yield
        self._ignoring_notifs = False

    def __init__(self, *args, **kwargs):
        """
        Abstract class that implements the model side of the adapter.
        For a fully implemented class check the corresponding widget toolkit
        modules. 

        This class handles only assignments to properties. For other
        kinds of setting (e.g. user-defined classes used as
        observable properties, containers, etc.) use other types of
        Adapters derived from this class.
        
        Controller is the controller creating this Adapter.
        Prop is either a :class:`~..models.properties.LabeledProperty` instance
        taken from the controller's model, or a 'dotted string' describing a
        property from the controller's model. E.g. it is possible to pass: 
        a.b.c which will then result in property c from attribute b from 
        attribute a from the controller's model to be observed.
        This way nested properties can also be handled.
        
        prop_{read, write} are two optional functions that apply
        custom modifications to the value of the property after
        reading and before setting it on the model (e.g. transforming a float
        to a string and vice-versa). Both take a value and must return a
        transformed value.
        
        value_error can be a function to be called when a ValueError exception
        occurs while trying to set an invalid value for the property on the
        model. The function will receive: the adapter, the property name and
        the value coming from the widget that offended the model.

        spurious controls if the adapter should receive spurious
        changes from the model (see spuriousness in class Observer for
        further information).
        """
        controller = args[0]
        prop = args[1]
        widget = args[2]

        prop_read = kwargs.pop("prop_read", None)
        prop_write = kwargs.pop("prop_write", None)
        value_error = kwargs.pop("value_error", None)
        spurious = kwargs.get("spurious", False)

        # First parse (optional) property strings:
        prop, self._model = self._parse_prop(prop, controller.model)

        # Call base __init__'s:
        super(ModelAdapter, self).__init__(*args, **kwargs)

        # Mode-side property handling:
        self._prop_read = not_none(prop_read, self._prop_read)
        self._prop_write = not_none(prop_write, self._prop_write)
        self._value_error = not_none(value_error, self._value_error)

        # Connect the model
        self._connect_model()
        return

    # ----------------------------------------------------------------------
    #  Model connecting & disconnecting:
    # ----------------------------------------------------------------------
    def _parse_prop(self, prop, model):
        """Parses (optional) prop strings for the given model"""
        if not isinstance(prop, LabeledProperty):
            parts = prop.split(".")
            if len(parts) > 1:
                # identifies the model
                models = parts[:-1]
                for name in models:
                    model = getattr(model, name)
                    if not isinstance(model, Model):
                        raise TypeError(
                            "Attribute '%s' was expected to be a " +
                            "Model, but found: '%s'" % (name, model))
                prop = getattr(type(model), parts[-1])
            else: prop = parts[0]
        return prop, model

    def _connect_model(self):
        """Used internally to connect the property into the model, and
        register self as a value observer for that property"""

        # prop is inside model?
        if not hasattr(self._model, self._prop.label):
            raise ValueError("Attribute '" + self._prop.label +
                             "' not found in model " + str(self._model))

        # is it observable?
        if self._prop.observable:
            self.observe(self._on_prop_changed, self._prop.label, assign=True)
            self.observe_model(self._model)

    def _on_prop_changed(self, *args, **kwargs):
        """Called by the observation code, when the value in the
        observed property is changed"""
        if not self._ignoring_notifs:
            self.update_widget()

    def _disconnect_model(self, model=None):
        # disconnects the model
        model = not_none(self._model, model)
        if model is not None:
            self.relieve_model(self._model)
            self._model = None

    # ----------------------------------------------------------------------
    #  Model-side reading and writing
    # ----------------------------------------------------------------------
    def _get_property_value(self):
        """Private method that returns the property value currently stored
        in the model, without transformations."""
        return getattr(self._model, self._prop.label)

    def _set_property_value(self, val):
        """Private method that sets the property value stored in the model,
        without transformations."""
        return setattr(self._model, self._prop.label, val)

    def _read_property(self, *args):
        """Returns the (possibly transformed) property value stored in the 
        model"""
        if self._prop_read: return self._prop_read(self._get_property_value(*args))
        return self._get_property_value(*args)

    def _write_property(self, value, *args):
        """Sets the value of property. The given value is transformed
        using the prop_write function passed to the constructor.
        An attempt at casting the value to the property type is also made."""
        raw_value = value
        try:
            # transform if needed:
            if self._prop_write: value = self._prop_write(value)

            # set the property, ignore updates:
            with self._ignore_notifications():
                self._set_property_value(value, *args)

        except ValueError:
            # let the user handle the error if he wants that:
            if self._value_error:
                self._value_error(self, self._prop.label, raw_value)
            else:
                raise

    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

from .base import Specimen
from .markers import ThresholdSelector, Marker, MineralScorer
from .statistics import Statistics

__all__ = [
    "Specimen",
    "Thresholdselector",
    "Marker",
    "MineralScorer",
    "Statistics"
]



#!/usr/bin/env python2

import sys, os
import subprocess
import re
import fileinput

def update_version(filename, version):
    for line in fileinput.input(filename, inplace=True):
        rexp = "__version__\s*=.*"
        line = re.sub(rexp, "__version__ = \"%s\"" % version, line),
        print(line[0], end='')

assert len(sys.argv) > 1, "You need to specify the version (e.g. 6.6.6)"
assert "v" not in sys.argv[1], "You need to specify the version (e.g. 6.6.6)"

update_version(os.path.abspath("pyxrd/__version.py"), sys.argv[1])
update_version(os.path.abspath("mvc/__version.py"), sys.argv[1])

print( subprocess.check_output(['git', 'add', 'pyxrd/__version.py']))
print( subprocess.check_output(['git', 'add', 'mvc/__version.py']))
print( subprocess.check_output(['git', 'commit', '-m', 'Version bump', '--allow-empty']))
print( subprocess.check_output(['git', 'tag', '-a', 'v%s' % sys.argv[1], '-m', 'v%s' % sys.argv[1]]))



# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import traceback
import sys
import importlib, pkgutil

from imp import find_module

import logging
logger = logging.getLogger(__name__)

"""
    This scans the module for submodules and imports them. This will
    trigger the registration of any refinement method classes 
    (i.e. RefineRun subclasses).
    
    Every RefineRun class is callable. 
    When calling a RefineRun sub-class, you should pass the RefineContext as the first
    argument, a stop signal, and an optional dict of options (see the class
    definitions for what options you can use). 
    Internally, this will set-up the class and then call its own `run()` method, 
    starting the refinement.  

    As such, to an external user, these 'classes' appear as simple functions.
"""


def import_submodules(package_name):
    """ Import all submodules of a module, recursively

    :param package_name: Package name
    :type package_name: str
    :rtype: dict[types.ModuleType]
    """
    package = sys.modules[package_name]

    modules = {}
    for _, name, _ in pkgutil.walk_packages(package.__path__):
        try:
            modules[name] = importlib.import_module(package_name + '.' + name)
        except:
            logger.warning("Could not import %s refinement method modules, are all dependencies installed? Error was:" % name)
            _, _, tb = sys.exc_info()
            traceback.print_tb(tb)
    return modules

__all__ = list(import_submodules(__name__).keys())





# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#  Copyright (C) 2005 by Roberto Cavada <roboogle@gmail.com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------


class DecoratorError (SyntaxError):
    """Used to report syntax errors occurring when decorators are
    used in models and observers."""
    pass


class TooManyCandidatesError (ValueError):
    """This class is used for distinguishing between a
    multiple candidates matched and no candidates matched. The
    latter is not necessarily an issue, as a missed match can
    be skipped when searching for a match for *all* the
    properties in the model (no params to adapt()), which may
    fail in one single view, as multiple views may be used to
    represent different parts of the model"""
    pass


class ViewError (ValueError):
    """General issue with view content"""
    pass




# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

import random

from mvc.models.properties import (
    LabeledProperty, ListProperty, IntegerChoiceProperty, BoolProperty,
    ReadOnlyMixin
)
from mvc.models import TreeNode

from pyxrd.generic.models.event_context_manager import EventContextManager
from pyxrd.generic.models.properties import InheritableMixin
from pyxrd.generic.models import ChildModel

from .refinables.mixins import RefinementValue, RefinementGroup
from .refinables.wrapper import RefinableWrapper
from .refine_method_manager import RefineMethodManager
from .refiner import Refiner

class Refinement(ChildModel):
    """
        A simple model that plugs onto the Mixture model. It provides
        the functionality related to refinement of parameters.
    """

    # MODEL INTEL:
    class Meta(ChildModel.Meta):
        store_id = "Refinement"

    mixture = property(ChildModel.parent.fget, ChildModel.parent.fset)

    #: Flag, True if after refinement plots should be generated of the parameter space
    make_psp_plots = BoolProperty(
        default=False, text="Make parameter space plots",
        tabular=False, visible=True, persistent=True
    )

    #: TreeNode containing the refinable properties
    refinables = ListProperty(
        default=None, text="Refinables",
        tabular=True, persistent=False, visible=True, 
        data_type=RefinableWrapper,
        cast_to=None, widget_type="object_tree_view"
    )

    #: A dict containing an instance of each refinement method
    refine_methods = None


    #: An integer describing which method to use for the refinement
    refine_method_index = IntegerChoiceProperty(
        default=0, text="Refinement method index",
        tabular=True, persistent=True, visible=True,
        choices={ key: method.name for key, method in RefineMethodManager.get_all_methods().items() }
    )

    #: A dict containing the current refinement options
    @LabeledProperty(
        default=None, text="Refine options",
        persistent=False, visible=False,
        mix_with=(ReadOnlyMixin,)
    )
    def refine_options(self):
        return self.get_refinement_method().get_options()
    
    #: A dict containing all refinement options
    @property
    def all_refine_options(self):
        return {
            method.index : method.get_options()
            for method in list(self.refine_methods.values())
        }

    def __init__(self, *args, **kwargs):
        my_kwargs = self.pop_kwargs(kwargs,
            "refine_method_index", "refine_method", "refine_options"
        )
        super(Refinement, self).__init__(*args, **kwargs)
        kwargs = my_kwargs

        # Setup the refinables treestore
        self.refinables = TreeNode()
        self.update_refinement_treestore()

        # Setup the refine methods
        try:
            self.refine_method_index = int(self.get_kwarg(kwargs, None, "refine_method_index", "refine_method"))
        except ValueError:
            self.refine_method_index = self.refine_method_index
            pass # ignore faulty values, these indices change from time to time.

        self.refine_methods = RefineMethodManager.initialize_methods(
            self.get_kwarg(kwargs, None, "refine_options")
        )

    # ------------------------------------------------------------
    #      Refiner methods
    # ------------------------------------------------------------
    def get_refiner(self):
        """
            This returns a Refiner object which can be used to refine the
            selected properties using the selected algorithm.
            Just call 'refine(stop)' on the returned object, with stop a
            threading.Event or multiprocessing.Event which you can use to stop
            the refinement before completion.
            The Refiner object also has a RefineHistory and RefineStatus object
            that can be used to track the status and history of the refinement.
        """ 
       
        return Refiner(
            method            = self.get_refinement_method(),
            data_callback     = lambda: self.mixture.data_object,
            refinables        = self.refinables,
            event_cmgr        = EventContextManager(self.mixture.needs_update, self.mixture.data_changed),
            metadata          = dict(
                phases          = self.mixture.phases,
                num_specimens   = len(self.mixture.specimens),
            )
        )

    # ------------------------------------------------------------
    #      Refinement Methods Management
    # ------------------------------------------------------------
    def get_refinement_method(self):
        """
            Returns the actual refinement method by translating the 
            `refine_method` attribute
        """
        return self.refine_methods[self.refine_method_index]

    # ------------------------------------------------------------
    #      Refinables Management
    # ------------------------------------------------------------
    # TODO set a restrict range attribute on the PropIntels, so we can use custom ranges for each property
    def auto_restrict(self):
        """
            Convenience function that restricts the selected properties 
            automatically by setting their minimum and maximum values.
        """
        with self.mixture.needs_update.hold():
            for node in self.refinables.iter_children():
                ref_prop = node.object
                if ref_prop.refine and ref_prop.refinable:
                    ref_prop.value_min = ref_prop.value * 0.8
                    ref_prop.value_max = ref_prop.value * 1.2

    def randomize(self):
        """
            Convenience function that randomize the selected properties.
            Respects the current minimum and maximum values.
            Executes an optimization after the randomization.
        """
        with self.mixture.data_changed.hold_and_emit():
            with self.mixture.needs_update.hold_and_emit():
                for node in self.refinables.iter_children():
                    ref_prop = node.object
                    if ref_prop.refine and ref_prop.refinable:
                        ref_prop.value = random.uniform(ref_prop.value_min, ref_prop.value_max)

    def update_refinement_treestore(self):
        """
            This creates a tree store with all refinable properties and their
            minimum, maximum and current value.
        """
        if self.parent is not None: # not linked so no valid phases!
            self.refinables.clear()

            def add_property(parent_node, obj, prop, is_grouper):
                rp = RefinableWrapper(obj=obj, prop=prop, parent=self.mixture, is_grouper=is_grouper)
                return parent_node.append(TreeNode(rp))

            def parse_attribute(obj, prop, root_node):
                """
                    obj: the object
                    attr: the attribute of obj or None if obj contains attributes
                    root_node: the root TreeNode new iters should be put under
                """
                if prop is not None:
                    if isinstance(prop, InheritableMixin):
                        value = prop.get_uninherited(obj)
                    else:
                        value = getattr(obj, prop.label)
                else:
                    value = obj

                if isinstance(value, RefinementValue): # AtomRelation and UnitCellProperty
                    new_node = add_property(root_node, value, prop, False)
                elif hasattr(value, "__iter__"): # List or similar
                    for new_obj in value:
                        parse_attribute(new_obj, None, root_node)
                elif isinstance(value, RefinementGroup): # Phase, Component, Probability
                    if len(value.refinables) > 0:
                        new_node = add_property(root_node, value, prop, True)
                        for prop in value.refinables:
                            parse_attribute(value, prop, new_node)
                else: # regular values
                    new_node = add_property(root_node, obj, prop, False)

            for phase in self.mixture.project.phases:
                if phase in self.mixture.phase_matrix:
                    parse_attribute(phase, None, self.refinables)


    pass # end of class



# coding=UTF-8
# ex:ts=4:sw=4:et=on
#  -------------------------------------------------------------------------
#  Copyright (C) 2014 by Mathijs Dumon <mathijs dot dumon at gmail dot com>
#
#  mvc is a framework derived from the original pygtkmvc framework
#  hosted at: <http://sourceforge.net/projects/pygtkmvc/>
#
#  mvc is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 2 of the License, or (at your option) any later version.
#
#  mvc is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor,
#  Boston, MA 02110, USA.
#  -------------------------------------------------------------------------

from .base import Observer

class TreeObserver(Observer):
    """
        An observer that wraps the in-instance changes of a Tree made by
        TreeNodes to on_inserted and on_deleted handlers.
    """

    _deleted = []

    def __init__(self, on_inserted, on_deleted, prop_name, on_deleted_before=None, model=None, spurious=False):
        super(TreeObserver, self).__init__(spurious=spurious)
        self.on_inserted = on_inserted
        self.on_deleted = on_deleted
        self.on_deleted_before = on_deleted_before

        self.observe(self.on_prop_mutation_before, prop_name, before=True)
        self.observe(self.on_prop_mutation_after, prop_name, after=True)
        self.observe_model(model)

    def on_prop_mutation_before(self, model, prop_name, info):
        if info.method_name == "on_grandchild_removed":
            self._deleted = [info.args[0], ]

        if info.method_name == "remove":
            self._deleted = [info.args[0], ]

        if callable(self.on_deleted_before):
            for old_item in self._deleted[::-1]:
                self.on_deleted_before(old_item)

    def on_prop_mutation_after(self, model, prop_name, info):
        if callable(self.on_deleted):
            for old_item in self._deleted[::-1]:
                self.on_deleted(old_item)
            self._deleted = []

        if info.method_name == "on_grandchild_removed":
            new_item = info.args[0]
            self.on_inserted(new_item)

        if info.method_name == "insert":
            new_item = info.args[1]
            self.on_inserted(new_item)

    pass # end of class


# coding=UTF-8
# ex:ts=4:sw=4:et=on

# Copyright (c) 2013, Mathijs Dumon
# All rights reserved.
# Complete license can be found in the LICENSE file.

import logging
logger = logging.getLogger(__name__)

import numpy as np

import matplotlib
import matplotlib.colors as mcolors
import matplotlib.transforms as transforms
from matplotlib.patches import FancyBboxPatch, Rectangle
from matplotlib.offsetbox import VPacker, HPacker, AnchoredOffsetbox, TextArea, AuxTransformBox
from matplotlib.text import Text

from pyxrd.data import settings
from pyxrd.calculations.math_tools import smooth, add_noise

from .draggables import DraggableMixin

def getattr_or_create(obj, attr, create):
    value = getattr(obj, attr, None)
    if value == None:
        class_type, args, kwargs = create
        value = class_type(*args, **kwargs)
    return value

def plot_marker_text(project, marker, offset, marker_scale, base_y, axes):
    """
        Plots a markers text using the given offset and scale
    """
    text = getattr(marker, "__plot_text", None)
    within_range = bool(
        project.axes_xlimit == 0 or
        (marker.position >= project.axes_xmin and
        marker.position <= project.axes_xmax)
    )
    if marker.visible and marker.style != "offset" and within_range:
        # prevent empty $$ from causing an error:
        save_label = marker.label.replace("$$", "")

        # Calculate position and set transform:
        x = float(marker.position) + float(marker.x_offset)
        if marker.top == 0: # relative to base
            y = base_y + (marker.top_offset + marker.y_offset) * marker_scale
            transform = axes.transData
        elif marker.top == 1: # top of plot
            y = settings.PLOT_TOP + float(marker.y_offset)
            transform = transforms.blended_transform_factory(axes.transData, axes.get_figure().transFigure)

        kws = dict(text=save_label,
                   x=x, y=y,
                   clip_on=False,
                   transform=transform,
                   horizontalalignment=marker.align, verticalalignment="center",
                   rotation=(90 - marker.angle), rotation_mode="anchor",
                   color=marker.color,
                   weight="heavy")

        if text:
            for key in kws: getattr(text, "set_%s" % key)(kws[key])
        else:
            text = Text(**kws)
        if not text in axes.get_children():
            axes.add_artist(text)
    elif text:
        try: text.remove()
        except: pass
    marker.__plot_text = text
    return text

def plot_marker_line(project, marker, offset, base_y, axes):
    """
        Plots a markers connector line using the given offset
    """
    line = getattr(marker, "__plot_line", None)
    within_range = bool(
        project.axes_xlimit == 0 or
        (marker.position >= project.axes_xmin and
        marker.position <= project.axes_xmax)
    )
    if marker.visible and within_range:
        # We need to strip away the units for comparison with
        # non-unitized bounds
        trans = transforms.blended_transform_factory(axes.transData, axes.transAxes)

        # Calculate top and bottom positions:
        ymin, ymax = axes.get_ybound()
        y = base_y
        y0 = (y - ymin) / (ymax - ymin)
        if marker.top == 0: # relative to base
            y1 = y0 + (marker.top_offset - ymin) / (ymax - ymin)
        elif marker.top == 1: # top of plot
            y1 = 1.0

        # If style is 'offset', re-calculate positions accordingly
        style = marker.style
        if style == "offset":
            style = "solid"
            y0 = (offset - ymin) / (ymax - ymin)
            y1 = y0 + (marker.y_offset - ymin) / (ymax - ymin)

        data = [y0, y1]

        if line:
            line.set_xdata(np.array([marker.position, marker.position]))
            line.set_ydata(np.array(data))
            line.set_transform(trans)
            line.set_color(marker.color)
            line.set_linestyle(style)
        else:
            line = matplotlib.lines.Line2D([marker.position, marker.position], data , transform=trans, color=marker.color, ls=style)
            line.y_isdata = False

        if not line in axes.get_lines():
            axes.add_line(line)
    elif line:
        try: line.remove()
        except: pass
    marker.__plot_line = line

def plot_markers(cc, project, specimen, marker_lbls, offset, scale, marker_scale, axes):
    """
        Plots a specimens markers using the given offset and scale
    """

    for marker in specimen.markers:
        base_y = 0
        if marker.base == 1:
            base_y = specimen.experimental_pattern.get_plotted_y_at_x(marker.position)
        elif marker.base == 2:
            base_y = specimen.calculated_pattern.get_plotted_y_at_x(marker.position)
        elif marker.base == 3:
            base_y = min(
                specimen.experimental_pattern.get_plotted_y_at_x(marker.position),
                specimen.calculated_pattern.get_plotted_y_at_x(marker.position)
            )
        elif marker.base == 4:
            base_y = max(
                specimen.experimental_pattern.get_plotted_y_at_x(marker.position),
                specimen.calculated_pattern.get_plotted_y_at_x(marker.position)
            )

        plot_marker_line(project, marker, offset, base_y, axes)
        text = plot_marker_text(project, marker, offset, marker_scale, base_y, axes)
        if text is not None:
            cc.register_artist(text, marker)
            marker_lbls.append((text, marker.base == 0, marker.y_offset))

def plot_hatches(project, specimen, offset, scale, axes):
    """
        Plots a specimens exclusion 'hatched' areas using the given offset and
        scale
    """
    # calculate the Y limits
    y0 = offset
    y1 = offset + max(specimen.max_display_y * scale, 1.0)

    # these are easier to just remove for now, not too expensive
    leftborder, hatch, rightborder = getattr(specimen, "__plot_hatches_artists", (None, None, None))
    if leftborder:
        try: leftborder.remove()
        except: pass
    if hatch:
        try: hatch.remove()
        except: pass
    if rightborder:
        try: rightborder.remove()
        except: pass

    # Create & add new hatches:
    for x0, x1 in zip(*specimen.exclusion_ranges.get_xy_data()):
        leftborder = axes.plot([x0, x0], [y0, y1], c=settings.EXCLUSION_LINES)
        axes.add_patch(Rectangle(
            (x0, y0), x1 - x0, y1 - y0,
            fill=True, hatch="/", linewidth=0,
            facecolor=settings.EXCLUSION_FOREG,
            edgecolor=settings.EXCLUSION_LINES)
        )
        rightborder = axes.plot([x1, x1], [y0, y1], c=settings.EXCLUSION_LINES)

def plot_label(specimen, labels, label_offset, plot_left, axes):
    text = getattr(specimen, "__plot_label_artist", None)

    # prevent empty $$ from causing an error:
    save_label = specimen.label.replace("$$", "")

    props = dict(
        text=save_label,
        x=plot_left - 0.05,
        y=label_offset,
        clip_on=False,
        horizontalalignment='right',
        verticalalignment='center',
        transform=transforms.blended_transform_factory(axes.get_figure().transFigure, axes.transData)
    )
    if text:
        for key in props: getattr(text, "set_%s" % key)(props[key])
    else:
        text = Text(**props)
    if not text in axes.get_children():
        axes.add_artist(text)
    labels.append(text)
    specimen.__plot_label_artist = text

def apply_transform(data, scale=1, offset=0, cap=0):
    data_x, data_y = data
    data_y = np.array(data_y) # make a copy
    if cap > 0:
        np.copyto(data_y, [cap], where=(data_y >= cap)) # copy the cap where values are larger then cap
    data_y = data_y * scale + offset # scale and offset the capped data
    return data_x, data_y

def plot_pattern(pattern, axes, scale=1, offset=0, cap=0, z_data=[None], **kwargs):
    
    if len(z_data) > 1 and pattern.data_y.size > 0:
        # Update the 2d image   
        colors = [mcolors.colorConverter.to_rgba(pattern.color, 0), mcolors.colorConverter.to_rgba(pattern.color, 1)]
        cmap = mcolors.LinearSegmentedColormap.from_list('mycmap', colors, N=100)
        
        angles = np.array(pattern.data_x, dtype=float)
        intensities = np.array(pattern.data_y[:,:len(z_data)], dtype=float) * scale + offset
        rh = np.array(list(map(float, z_data)), dtype=float)
        
        rh = rh * scale + offset
        
        cols = intensities.shape[0]
        rows = intensities.shape[1]
        X = np.ones(shape=(rows, cols), dtype=float) * angles #ANGLE
        Y = (np.ones(shape=(cols, rows), dtype=float) * rh).transpose() #RH
        Z = intensities.transpose()
        axes.pcolormesh(X,Y,Z, cmap=cmap, shading='flat', edgecolors='None', rasterized=False)       
        
    else:
        # setup or update the line
        line = getattr_or_create(pattern, "__plot_line", (matplotlib.lines.Line2D, ([], []), {}))
    
        if kwargs:
            line.update(kwargs)
        line.update(dict(
            data=apply_transform(pattern.get_xy_data(), scale=scale, offset=offset, cap=cap),
            color=pattern.color,
            linewidth=pattern.lw,
            ls=getattr(pattern, "ls", "-"),
            marker=getattr(pattern, "marker", "")
        ))
        if not line in axes.get_lines():
            axes.add_line(line)
        pattern.__plot_line = line

def make_draggable(artist, drag_x_handler=None, drag_y_handler=None):
    if artist != None:
        draggable = getattr(artist, "__draggable", None)
        if draggable == None:
            draggable = DraggableMixin(artist, drag_x_handler, drag_y_handler)
        else:
            draggable.update(artist, drag_x_handler, drag_y_handler)
        artist.__draggable = draggable

def plot_specimen(cc, project, specimen, labels, marker_lbls, label_offset, plot_left,
        offset, scale, marker_scale, axes):
    """
        Plots a specimens patterns, markers and hatches using the given
        offset and scale
    """
    # Plot the patterns;

    z_data = specimen.get_z_list()

    if specimen.display_experimental:
        pattern = specimen.experimental_pattern

        # plot the experimental pattern:
        plot_pattern(pattern, axes, scale=scale, offset=offset, cap=pattern.cap_value, z_data=z_data)
        #make_draggable(getattr(pattern, "__plot_line", None), drag_y_handler=specimen.on_pattern_dragged)

        # get some common data for the next lines:
        x_data, y_data = pattern.get_xy_data()
        xmin, xmax = (np.min(x_data), np.max(x_data)) if x_data.size > 0 else (0, 0)
        ymin, ymax = (np.min(y_data), np.max(y_data)) if y_data.size > 0 else (0, 0)

        ########################################################################
        # plot the background pattern:
        bg_line = getattr_or_create(pattern, "__plot_bg_line", (matplotlib.lines.Line2D, ([], []), dict(c="#660099", lw="2", zorder=10)))
        if pattern.bg_type == 0 and pattern.bg_position != 0.0:
            bg_line.update(dict(
                data=apply_transform(([xmin, xmax], [pattern.bg_position, pattern.bg_position]), scale=scale, offset=offset),
                visible=True
            ))
        elif pattern.bg_type == 1 and pattern.bg_pattern is not None:
            bg_line.update(dict(
                data=apply_transform((x_data, (pattern.bg_pattern * pattern.bg_scale) + pattern.bg_position), scale=scale, offset=offset),
                visible=True
            ))
        else:
            bg_line.update(dict(
                data=([], []),
                visible=True
            ))

        if bg_line.get_visible() and not bg_line in axes.get_lines():
            axes.add_line(bg_line)
        elif not bg_line.get_visible():
            try: bg_line.remove()
            except: pass
        pattern.__plot_bg_line = bg_line
        ########################################################################

        ########################################################################
        # plot the smooth pattern:
        smooth_line = getattr_or_create(pattern, "__plot_smooth_line", (matplotlib.lines.Line2D, ([], []), dict(c="#660099", lw="2", zorder=10)))

        if int(pattern.smooth_degree) > 1:
            data = x_data, smooth(y_data, pattern.smooth_degree)
        else:
            data = [], []
        smooth_line.update(dict(
            data=apply_transform(data, scale=scale, offset=offset),
            visible=bool(pattern.smooth_degree > 1)
        ))
        if smooth_line.get_visible() and not smooth_line in axes.get_lines():
            axes.add_line(smooth_line)
        elif not smooth_line.get_visible():
            try: smooth_line.remove()
            except: pass
        pattern.__plot_smooth_line = smooth_line
        ########################################################################

        ########################################################################
        # plot the noisified pattern:
        noise_line = getattr_or_create(pattern, "__plot_noise_line", (matplotlib.lines.Line2D, ([], []), dict(c="#660099", lw="2", zorder=10)))


        if pattern.noise_fraction > 0.0:
            data = x_data, add_noise(y_data, pattern.noise_fraction)
        else:
            data = [], []
        noise_line.update(dict(
            data=apply_transform(data, scale=scale, offset=offset),
            visible=bool(pattern.noise_fraction > 0.0)
        ))
        if noise_line.get_visible() and not noise_line in axes.get_lines():
            axes.add_line(noise_line)
        elif not noise_line.get_visible():
            try: noise_line.remove()
            except: pass
        pattern.__plot_noise_line = noise_line
        ########################################################################

        ########################################################################
        # plot the shift & reference lines:
        shifted_line = getattr_or_create(pattern, "__plot_shifted_line", (matplotlib.lines.Line2D, ([], []), dict(c="#660099", lw="2", zorder=10)))
        reference_line = getattr_or_create(pattern, "__plot_reference_line", (matplotlib.lines.Line2D, ([], []), dict(c="#660099", lw="2", ls="--", zorder=10)))

        if pattern.shift_value != 0.0:
            shifted_line.update(dict(
                data=apply_transform((x_data - pattern._shift_value, y_data.copy()), scale=scale, offset=offset),
                visible=True
            ))
            position = specimen.goniometer.get_2t_from_nm(pattern.shift_position)
            reference_line.update(dict(
                data=apply_transform(([position, position], [0, ymax]), scale=scale, offset=offset),
                visible=True
            ))
            if not shifted_line in axes.get_lines():
                axes.add_line(shifted_line)
            if not reference_line in axes.get_lines():
                axes.add_line(reference_line)
        else:
            shifted_line.set_data([], [])
            shifted_line.set_visible(False)
            try: shifted_line.remove()
            except: pass
            reference_line.set_data([], [])
            reference_line.set_visible(False)
            try: reference_line.remove()
            except: pass
        pattern.__plot_shifted_line = shifted_line
        pattern.__plot_reference_line = reference_line
        ########################################################################

        ########################################################################
        # plot the pattern after peak stripping:
        stripped_line = getattr_or_create(pattern, "__plot_stripped_line", (matplotlib.lines.Line2D, ([], []), dict(c="#660099", lw="1", zorder=10)))


        if pattern.strip_startx != 0.0 and pattern.strip_endx != 0.0:
            strip_xdata, strip_ydata = pattern.stripped_pattern
            stripped_line.update(dict(
                data=apply_transform((strip_xdata.copy(), strip_ydata.copy()), scale=scale, offset=offset),
                visible=True
            ))
            if not stripped_line in axes.get_lines():
                axes.add_line(stripped_line)
        else:
            stripped_line.set_data([], [])
            stripped_line.set_visible(False)
            try: stripped_line.remove()
            except: pass

        pattern.__plot_stripped_line = stripped_line
        ########################################################################

        ########################################################################
        # plot the pattern after peak stripping:
        artists = getattr(specimen, "__plot_peak_area", None)
        peak_area, peak_fwhm_line = artists if artists is not None else (None, None)
        if peak_area is not None and peak_area in axes.get_children():
            peak_area.remove()
        if peak_fwhm_line is not None and peak_fwhm_line in axes.get_children():
            peak_fwhm_line.remove()
        if pattern.peak_startx != 0.0 and pattern.peak_endx != 0.0 and pattern.peak_properties_pattern is not None:
            peak_xdata, peak_bg, peak_ydata, peak_rootsx, peak_rootsy = pattern.peak_properties_pattern
            _, peak_bg = apply_transform((peak_xdata.copy(), peak_bg.copy()), scale=scale, offset=offset)
            peak_xdata, peak_ydata = apply_transform((peak_xdata.copy(), peak_ydata.copy()), scale=scale, offset=offset)
            peak_rootsx, peak_rootsy = apply_transform((peak_rootsx.copy(), peak_rootsy.copy()), scale=scale, offset=offset)
            peak_area = axes.fill_between(peak_xdata, peak_bg, peak_ydata, interpolate=True, facecolor="#660099", zorder=10)
            peak_fwhm_line = axes.plot(peak_rootsx, peak_rootsy, color="#330033", zorder=11)
        setattr(specimen, "__plot_peak_area", (peak_area, peak_fwhm_line))

    if specimen.display_calculated:
        pattern = specimen.calculated_pattern
        plot_pattern(pattern, axes, scale=scale, offset=offset, z_data=z_data)
        #if not specimen.display_experimental:
        #    make_draggable(getattr(pattern, "__plot_line", None), drag_y_handler=specimen.on_pattern_dragged)

        # setup or update the calculated lines (phases)
        if specimen.display_phases:
            phase_lines = getattr(specimen, "__plot_phase_lines", [])

            # Clear previous phase lines:
            for phase_line in phase_lines:
                if phase_line in axes.get_lines():
                    axes.remove_line(phase_line)

            # Update & add phase lines:
            for i in range(2, pattern.num_columns):
                phase_data = pattern.get_xy_data(i)
                # Get the line object or create it:
                try:
                    phase_line = phase_lines[i - 2]
                except IndexError:
                    phase_line = matplotlib.lines.Line2D(*phase_data)
                    phase_lines.append(phase_line)
                # Get the phase color or use a default color:
                try:
                    phase_color = pattern.phase_colors[i - 2]
                except IndexError:
                    phase_color = pattern.color
                # Update the line object properties:
                phase_line.update(dict(
                    data=apply_transform(phase_data, scale=scale, offset=offset),
                    color=phase_color,
                    linewidth=pattern.lw
                ))

                # Add to axes:
                axes.add_line(phase_line)

            specimen.__plot_phase_lines = phase_lines

    # mineral preview sticks
    if hasattr(specimen, "mineral_preview") and specimen.mineral_preview is not None:
        name, peaks = specimen.mineral_preview
        lines = getattr(specimen, "__plot_mineral_preview", [])
        for line in lines:
            try: line.remove()
            except: pass

        lines = []
        for position, intensity in peaks:
            position = specimen.goniometer.get_2t_from_nm(position / 10.)
            intensity /= 100.

            trans = transforms.blended_transform_factory(axes.transData, axes.transAxes)
            ymin, ymax = axes.get_ybound()
            style = "solid"
            color = "#FF00FF"
            y0 = (offset - ymin) / (ymax - ymin)
            y1 = y0 + (intensity - ymin) / (ymax - ymin)
            line = matplotlib.lines.Line2D(
                [position, position], [y0, y1],
                transform=trans, color=color, ls=style
            )
            axes.add_line(line)
            lines.append(line)
        setattr(specimen, "__plot_mineral_preview", lines)


    # exclusion ranges;
    plot_hatches(project, specimen, offset, scale, axes)
    # markers;
    plot_markers(cc, project, specimen, marker_lbls, offset, scale, marker_scale, axes)
    # & label:
    plot_label(specimen, labels, label_offset, plot_left, axes)
    #make_draggable(getattr(specimen, "__plot_label_artist", None), drag_y_handler=project.on_label_dragged)

def plot_statistics(project, specimen, spec_scale, stats_y_pos, stats_height, axes):

    # Scales & shifts the pattern so the zero line plots in the middle.
    def plot_pattern_middle(pattern, axes, height, vscale, offset, **kwargs):
        """
            Height is the fraction of the plot reserved for the residual pattern
            vscale is a user scaling factor applied to the residual pattern
            offset is the offset of the residual pattern position from the x-axis of the plot
        """
        # Offset to the middle of the available space:
        offset = offset + 0.5 * height
        # If the intensity difference is smaller then the space available, don't scale
        scale = spec_scale * 0.5 * vscale
        plot_pattern(pattern, axes, scale=scale, offset=offset, **kwargs)

    if specimen.display_residuals and specimen.statistics.residual_pattern is not None:
        plot_pattern_middle(
            specimen.statistics.residual_pattern,
            axes, height=stats_height,
            vscale=specimen.display_residual_scale,
            offset=stats_y_pos, alpha=0.75
        )
    if specimen.display_derivatives:
        for pattern in (
                specimen.statistics.der_residual_pattern,
                specimen.statistics.der_exp_pattern,
                specimen.statistics.der_calc_pattern):
            if pattern is not None:
                plot_pattern_middle(
                    pattern, axes, height=stats_height,
                    vscale=specimen.display_residual_scale,
                    offset=stats_y_pos, alpha=0.65
                )

def plot_specimens(axes, pos_setup, cc, project, specimens):
    """
        Plots multiple specimens within the context of a project
    """

    base_offset = project.display_plot_offset
    base_height = 1.0
    label_offset = project.display_label_pos

    scale, scale_unit = project.get_scale_factor()

    labels, marker_lbls = list(), list()
    current_y_pos = 0
    lbl_y_offset = 0
    group_counter = 0 # 'group by' specimen counter

    ylim = 0 # used to keep track of maximum y-value, for a tight y-axis

    for _, specimen in enumerate(specimens):

        spec_max_display_y = float(specimen.max_display_y)

        # single specimen normalization:
        if project.axes_ynormalize == 1:
            scale = (1.0 / spec_max_display_y) if spec_max_display_y != 0.0 else 1.0

        spec_y_offset = specimen.display_vshift * scale_unit
        spec_y_pos = current_y_pos * scale_unit + spec_y_offset
        spec_alloc_height = base_height * scale_unit
        spec_reqst_height = spec_alloc_height * specimen.display_vscale

        lbl_y_offset = (label_offset + specimen.display_vshift) * scale_unit
        lbl_y_pos = current_y_pos * scale_unit + lbl_y_offset

        # For the y-limit we do not add the specimens vscale or vshift:
        ylim = current_y_pos * scale_unit + spec_alloc_height

        # Specimen scale = global scale, adjusted by specimen vscale
        spec_scale = scale * specimen.display_vscale

        # when statistics are plotted,
        # 65% of the height goes to the actual specimen plots
        # 35% goes to the statistics plot:
        if project.layout_mode == "FULL" and (specimen.display_residuals or specimen.display_derivatives):
            stats_y_pos = spec_y_pos
            stats_height = 0.35 * spec_reqst_height

            spec_y_pos = spec_y_pos + stats_height
            spec_scale = spec_scale * 0.65

            plot_statistics(
                project, specimen, spec_scale,
                stats_y_pos, stats_height,
                axes
            )

        plot_specimen(
            cc, project, specimen, labels, marker_lbls,
            lbl_y_pos, pos_setup.left, spec_y_pos, spec_scale, scale_unit,
            axes
        )

        # increase offsets:
        group_counter += 1
        if group_counter >= project.display_group_by:
            group_counter = 0
            current_y_pos += base_offset

    axes.set_ylim(top=ylim)

    return labels, marker_lbls

def plot_mixtures(axes, project, mixtures):
    legend = getattr(project, "__plot_mixture_legend", None)
    if legend:
        try: legend.remove()
        except ValueError: pass

    figure = axes.get_figure()
    trans = figure.transFigure

    def create_rect_patch(ec="#000000", fc=None):
        _box = AuxTransformBox(transforms.IdentityTransform())
        rect = FancyBboxPatch(
            xy=(0, 0),
            width=0.02,
            height=0.02,
            boxstyle='square',
            ec=ec,
            fc=fc,
            mutation_scale=14, # font size
            transform=trans,
            alpha=1.0 if (ec is not None or fc is not None) else 0.0
        )
        _box.add_artist(rect)
        return _box

    legends = []
    for mixture in mixtures:
        legend_items = []

        # Add title:
        title = TextArea(mixture.name)
        title_children = [create_rect_patch(ec=None) for spec in mixture.specimens]
        title_children.insert(0, title)
        title_box = HPacker(children=title_children, align="center", pad=5, sep=3)
        legend_items.append(title_box)

        # Add phase labels & boxes
        for i, (phase, fraction) in enumerate(zip(mixture.phases, mixture.fractions)):
            label_text = "{}: {:>5.1f}".format(phase, fraction * 100.0)
            label = TextArea(label_text)
            phase_children = [
                create_rect_patch(fc=phase.display_color)
                for phase in mixture.phase_matrix[:, i].flat if phase is not None
            ]
            phase_children.insert(0, label)
            legend_items.append(
                HPacker(children=phase_children, align="center", pad=0, sep=3)
            )

        # Add created legend to the list:
        legends.append(
            VPacker(children=legend_items, align="right", pad=0, sep=3)
        )

    # Only add this if there's something to add!
    if legends:
        # Pack legends & plot:
        legend = AnchoredOffsetbox(
            loc=1,
            pad=0.1,
            borderpad=0.1,
            frameon=False,
            child=VPacker(children=legends, align="right", pad=0, sep=5)
        )

        axes.add_artist(legend)
        setattr(project, "__plot_mixture_legend", legend)


